#
# Copyright 2018-2023- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Custom database for store Persistent objects."""

import datetime
import hashlib
import importlib
import io
import json
from enum import Enum
from pathlib import Path
from types import BuiltinFunctionType, FunctionType
from typing import Any, Dict, List, Optional, Union, cast
from uuid import uuid4

import deal
import persistent
import zstandard as zstd
from BTrees.Length import Length
from BTrees.OOBTree import BTree, OOBucket, OOSet, OOTreeSet
from persistent import GHOST, UPTODATE
from persistent.interfaces import IPickleCache
from zc.relation.catalog import Catalog
from ZODB.utils import z64
from zope.interface import implementer
from zope.interface.interface import InterfaceClass

from renku.core import errors
from renku.domain_model.project import Project
from renku.infrastructure.immutable import Immutable
from renku.infrastructure.persistent import Persistent

OID_TYPE = str
TYPE_TYPE = "type"
FUNCTION_TYPE = "function"
REFERENCE_TYPE = "reference"
SET_TYPE = "set"
FROZEN_SET_TYPE = "frozenset"
MARKER = object()
"""These are used as _p_serial to mark if an object was read from storage or is new"""

NEW = z64  # NOTE: Do not change this value since this is the default when a Persistent object is created
PERSISTED = b"1" * 8


def _is_module_allowed(module_name: str, type_name: str):
    """Checks whether it is allowed to import from the given module for security purposes.

    Args:
        module_name(str): The module name to check.
        type_name(str): The type within the module to check.

    Raises:
        TypeError: If the type is now allowed in the database.
    """

    if module_name not in ["BTrees", "builtins", "datetime", "persistent", "renku", "zc", "zope", "deal"]:
        raise TypeError(f"Objects of type '{type_name}' are not allowed")


def get_type_name(object) -> Optional[str]:
    """Return fully-qualified object's type name.

    Args:
        object:  The object to get the type name for.

    Returns:
        Optional[str]: The fully qualified type name.

    """
    if object is None:
        return None

    object_type = object if isinstance(object, type) else type(object)
    return f"{object_type.__module__}.{object_type.__qualname__}"


def get_class(type_name: Optional[str]) -> Optional[type]:
    """Return the class for a fully-qualified type name.

    Args:
        type_name(Optional[str]): The name of the class to get.

    Returns:
        Optional[type]: The class.

    """
    if type_name is None:
        return None

    components = type_name.split(".")
    module_name = components[0]

    _is_module_allowed(module_name, type_name)

    module = __import__(module_name)

    return get_attribute(module, components[1:])


def get_attribute(object, name: Union[List[str], str]):
    """Return an attribute of an object.

    Args:
        object: The object to get an attribute on.
        name(Union[List[str], str): The name of the attribute to get.

    Returns:
        The value of the attribute.
    """
    import sys

    components = name.split(".") if isinstance(name, str) else name

    def _module_name(o):
        return o.__module__ if hasattr(o, "__module__") else o.__name__

    module_name = _module_name(object)
    root_module_name = module_name.split(".")[0]

    for component in components:
        module_name = _module_name(object)
        if not hasattr(object, component) and f"{module_name}.{component}" not in sys.modules:
            try:
                _is_module_allowed(root_module_name, object.__name__)
                object = importlib.import_module(f".{component}", package=module_name)
                continue
            except ModuleNotFoundError:
                pass

        object = getattr(object, component)

    return object


class RenkuOOBTree(BTree):
    """Customize ``BTrees.OOBTree.BTree`` implementation."""

    max_leaf_size = 1000
    max_internal_size = 2000


class Database:
    """The Metadata Object Database.

    This class is equivalent to a ``persistent.DataManager`` and implements
    the ``persistent.interfaces.IPersistentDataManager`` interface.
    """

    ROOT_OID = "root"

    def __init__(self, storage):
        self._storage: Storage = storage
        self._cache = Cache()
        # The pre-cache is used by get to avoid infinite loops when objects load their state
        self._pre_cache: Dict[OID_TYPE, persistent.Persistent] = {}
        # Objects added explicitly by add() or when serializing other objects. After commit they are moved to _cache.
        self._objects_to_commit: Dict[OID_TYPE, persistent.Persistent] = {}
        self._reader: ObjectReader = ObjectReader(database=self)
        self._writer: ObjectWriter = ObjectWriter(database=self)
        self._root: RenkuOOBTree

        self._initialize_root()

    @classmethod
    def from_path(cls, path: Union[Path, str]) -> "Database":
        """Create a Storage and Database using the given path.

        Args:
            path(Union[pathlib.Path, str]): The path of the database.

        Returns:
            The database object.
        """
        storage = Storage(path)
        return Database(storage=storage)

    @staticmethod
    def generate_oid(object: persistent.Persistent) -> OID_TYPE:
        """Generate an ``oid`` for a ``persistent.Persistent`` object based on its id.

        Args:
            object(persistent.Persistent): The object to create an oid for.

        Returns:
            An oid for the object.
        """
        oid = getattr(object, "_p_oid")
        if oid:
            assert isinstance(oid, OID_TYPE)
            return oid

        id: Optional[str] = getattr(object, "id", None) or getattr(object, "_id", None)
        if id:
            return Database.hash_id(id)

        return Database.new_oid()

    @staticmethod
    def hash_id(id: str) -> OID_TYPE:
        """Return ``oid`` from id.

        Args:
            id(str): The id to hash.

        Returns:
            OID_TYPE: The hashed id.
        """
        return hashlib.sha3_256(id.encode("utf-8")).hexdigest()

    @staticmethod
    def new_oid():
        """Generate a random ``oid``."""
        return f"{uuid4().hex}{uuid4().hex}"

    @staticmethod
    def _get_filename_from_oid(oid: OID_TYPE) -> str:
        return oid.lower()

    def __getitem__(self, key) -> "Index":
        return self._root[key]

    def clear(self):
        """Remove all objects and clear all caches. Objects won't be deleted in the storage."""
        self._cache.clear()
        self._pre_cache.clear()
        self._objects_to_commit.clear()
        # NOTE: Clear root at the end because it will be added to _objects_to_commit when `register` is called.
        self._root.clear()

    def _initialize_root(self):
        """Initialize root object."""
        try:
            self._root = cast(RenkuOOBTree, self.get(Database.ROOT_OID))
        except errors.ObjectNotFoundError:
            self._root = RenkuOOBTree()
            self._root._p_oid = Database.ROOT_OID
            self.register(self._root)

    def add_index(
        self, name: str, object_type: type, attribute: Optional[str] = None, key_type: Optional[type] = None
    ) -> "Index":
        """Add an index.

        Args:
            name(str): The name of the index.
            object_type(type): The type contained within the index.
            attribute(str, optional): The attribute of the contained object to create a key from (Default value = None).
            key_type(type, optional): The type of the key (Default value = None).

        Returns:
            Index: The created ``Index`` object.
        """
        assert name not in self._root, f"Index or object already exists: '{name}'"

        index = Index(name=name, object_type=object_type, attribute=attribute, key_type=key_type)
        index._p_jar = self

        self._root[name] = index

        return index

    def add_root_object(self, name: str, obj: Persistent):
        """Add an object to the DB root.

        Args:
            name(str): The key of the object.
            obj(Persistent): The object to store.
        """
        assert name not in self._root, f"Index or object already exists: '{name}'"

        obj._p_jar = self
        obj._p_oid = name

        self._root[name] = obj

    def add(self, object: persistent.Persistent, oid: OID_TYPE):
        """Add a new object to the database.

        NOTE: Normally, we add objects to indexes but this method adds objects directly to Dataset's root. Use it only
        for singleton objects that have no Index defined for them (e.g. Project).

        Args:
            object(persistent.Persistent): The object to add.
            oid(OID_TYPE, optional): The oid for the object (Default value = None).
        """
        assert not oid or isinstance(oid, OID_TYPE), f"Invalid oid type: '{type(oid)}'"
        object._p_oid = oid

        self.register(object)

    def register(self, object: persistent.Persistent):
        """Register a persistent.Persistent object to be stored.

        NOTE: When a persistent.Persistent object is changed it calls this method.

        Args:
            object(persistent.Persistent): The object to register with the database.
        """
        assert isinstance(object, persistent.Persistent), f"Cannot add non-Persistent object: '{object}'"

        if object._p_oid is None:
            object._p_oid = self.generate_oid(object)
        elif isinstance(object, Persistent):
            # NOTE: A safety-net to make sure that all objects have correct p_oid
            id = getattr(object, "id")
            expected_oid = Database.hash_id(id)
            actual_oid = object._p_oid
            assert actual_oid == expected_oid, f"Object has wrong oid: {actual_oid} != {expected_oid}"

        object._p_jar = self
        # object._p_serial = NEW
        self._objects_to_commit[object._p_oid] = object

    def get(self, oid: OID_TYPE) -> persistent.Persistent:
        """Get the object by ``oid``.

        Args:
            oid(OID_TYPE): The oid of the object to get.

        Returns:
            persistent.Persistent: The object.
        """
        if oid != Database.ROOT_OID and oid in self._root:  # NOTE: Avoid looping if getting "root"
            return self._root[oid]
        object = self.get_cached(oid)
        if object is not None:
            return object

        object = self.get_from_path(path=self._get_filename_from_oid(oid))

        if isinstance(object, Persistent):
            object.freeze()

        # NOTE: Avoid infinite loop if object tries to load its state before it is added to the cache
        self._pre_cache[oid] = object
        self._cache[oid] = object
        self._pre_cache.pop(oid)

        return object

    def get_from_path(
        self, path: str, absolute: bool = False, override_type: Optional[str] = None
    ) -> persistent.Persistent:
        """Load a database object from a path.

        Args:
            path(str): Path of the database object.
            absolute(bool): Whether the path is absolute or a filename inside the database (Default value = False).
            override_type(Optional[str]): load object as a different type than what is set inside `renku_data_type`
                (Default value = None).
        Returns:
            persistent.Persistent: The object.
        """
        deal.disable(warn=False)
        data = self._storage.load(filename=path, absolute=absolute)
        if override_type is not None:
            if "@renku_data_type" not in data:
                raise errors.IncompatibleParametersError("Cannot override type on found data.")

            data["@renku_data_type"] = override_type
        object = self._reader.deserialize(data)
        object._p_changed = 0
        object._p_serial = PERSISTED
        deal.enable()
        return object

    def get_by_id(self, id: str) -> persistent.Persistent:
        """Return an object by its id.

        Args:
            id(str): The id to look up.

        Returns:
            persistent.Persistent: The object with the given id.
        """
        oid = Database.hash_id(id)
        return self.get(oid)

    def get_cached(self, oid: OID_TYPE) -> Optional[persistent.Persistent]:
        """Return an object if it is in the cache or will be committed.

        Args:
            oid(OID_TYPE): The id of the object to look up.

        Returns:
            Optional[persistent.Persistent]: The cached object.
        """
        object = self._cache.get(oid)
        if object is not None:
            return object

        object = self._pre_cache.get(oid)
        if object is not None:
            return object

        object = self._objects_to_commit.get(oid)
        if object is not None:
            return object

        return None

    def remove_root_object(self, name: str) -> None:
        """Remove a root object from the database.

        Args:
            name(str): The name of the root object to remove.
        """
        assert name in self._root, f"Index or object doesn't exist in root: '{name}'"

        obj = self.get(name)
        self.remove_from_cache(obj)

        del self._root[name]

    def new_ghost(self, oid: OID_TYPE, object: persistent.Persistent):
        """Create a new ghost object.

        Args:
            oid(OID_TYPE): The oid of the new ghost object.
            object(persistent.Persistent): The object to create a new ghost entry for.
        """
        object._p_jar = self
        self._cache.new_ghost(oid, object)

    def setstate(self, object: persistent.Persistent):
        """Load the state for a ghost object.

        Args:
            object(persistent.Persistent): The object to set the state on.
        """
        deal.disable(warn=False)
        data = self._storage.load(filename=self._get_filename_from_oid(object._p_oid))
        self._reader.set_ghost_state(object, data)
        object._p_serial = PERSISTED
        if isinstance(object, Persistent):
            object.freeze()
        deal.enable()

    def commit(self):
        """Commit modified and new objects."""
        while self._objects_to_commit:
            _, object = self._objects_to_commit.popitem()
            if object._p_changed or object._p_serial == NEW:
                self._store_object(object)

    def _store_object(self, object: persistent.Persistent):
        data = self._writer.serialize(object)
        compress = False if isinstance(object, (Catalog, RenkuOOBTree, OOBucket, Project, Index)) else True
        self._storage.store(filename=self._get_filename_from_oid(object._p_oid), data=data, compress=compress)

        self._cache[object._p_oid] = object

        object._p_changed = 0  # NOTE: transition from changed to up-to-date
        object._p_serial = PERSISTED

    def persist_to_path(self, object: persistent.Persistent, path: Path):
        """Store an object to path."""
        data = self._writer.serialize(object)
        compress = False if isinstance(object, (Catalog, RenkuOOBTree, OOBucket, Project, Index)) else True
        self._storage.store(filename=str(path), data=data, compress=compress, absolute=True)

    def remove_from_cache(self, object: persistent.Persistent):
        """Remove an object from cache.

        Args:
            object(persistent.Persistent): The object to remove.
        """
        oid = object._p_oid

        def remove_from(cache):
            existing_entry = cache.get(oid)
            if existing_entry is object:
                cache.pop(oid)

        remove_from(self._cache)
        remove_from(self._pre_cache)
        remove_from(self._objects_to_commit)

    def readCurrent(self, object):
        """We don't use this method but some Persistent logic require its existence.

        Args:
            object: The object to read.
        """
        assert object._p_jar is self
        assert object._p_oid is not None

    def oldstate(self, object, tid):
        """See ``persistent.interfaces.IPersistentDataManager::oldstate``."""
        raise NotImplementedError


@implementer(IPickleCache)
class Cache:
    """Database ``Cache``."""

    def __init__(self):
        self._entries = {}

    def __len__(self):
        return len(self._entries)

    def __getitem__(self, oid):
        assert isinstance(oid, OID_TYPE), f"Invalid oid type: '{type(oid)}'"
        return self._entries[oid]

    def __setitem__(self, oid, object):
        assert isinstance(object, persistent.Persistent), f"Cannot cache non-Persistent objects: '{object}'"
        assert isinstance(oid, OID_TYPE), f"Invalid oid type: '{type(oid)}'"

        assert object._p_jar is not None, "Cached object jar missing"
        assert oid == object._p_oid, f"Cache key does not match oid: {oid} != {object._p_oid}"

        if oid in self._entries:
            existing_data = self.get(oid)
            if existing_data is not object:
                raise ValueError(f"The same oid exists: {existing_data} != {object}")

        self._entries[oid] = object

    def __delitem__(self, oid):
        assert isinstance(oid, OID_TYPE), f"Invalid oid type: '{type(oid)}'"
        self._entries.pop(oid)

    def clear(self):
        """Remove all entries."""
        self._entries.clear()

    def pop(self, oid, default=MARKER):
        """Remove and return an object.

        Args:
            oid: The oid of the object to remove from the cache.
            default: Default value to return (Default value = MARKER).
        Raises:
            KeyError: If object wasn't found and no default was given.
        Returns:
            The removed object or the default value if it doesn't exist.
        """
        return self._entries.pop(oid) if default is MARKER else self._entries.pop(oid, default)

    def get(self, oid, default=None):
        """See ``IPickleCache``.

        Args:
            oid: The oid of the object to get.
            default: Default value to return if object wasn't found (Default value = None).

        Returns:
            The object or default value if the object wasn't found.
        """
        assert isinstance(oid, OID_TYPE), f"Invalid oid type: '{type(oid)}'"
        return self._entries.get(oid, default)

    def new_ghost(self, oid, object):
        """See ``IPickleCache``."""
        assert object._p_oid is None, f"Object already has an oid: {object}"
        assert object._p_jar is not None, f"Object does not have a jar: {object}"
        assert oid not in self._entries, f"Duplicate oid: {oid}"

        object._p_oid = oid
        if object._p_state != GHOST:
            object._p_invalidate()

        self[oid] = object


class Index(persistent.Persistent):
    """Database index."""

    def __init__(self, *, name: str, object_type, attribute: Optional[str], key_type=None):
        """Create an index where keys are extracted using ``attribute`` from an object or a key.

        Args:
            name (str): Index's name.
            object_type: Type of objects that the index points to.
            attribute (Optional[str], optional): Name of an attribute to be used to automatically generate a key
                (e.g. `entity.path`).
            key_type: Type of keys. If not None then a key must be provided when updating the index
                (Default value = None).
        """
        assert name == name.lower(), f"Index name must be all lowercase: '{name}'."

        super().__init__()

        self._p_oid = f"{name}-index"
        self._name: str = name
        self._object_type = object_type
        self._key_type = key_type
        self._attribute: Optional[str] = attribute
        self._entries: RenkuOOBTree = RenkuOOBTree()
        self._entries._p_oid = name

    def __len__(self):
        return len(self._entries)

    def __contains__(self, key):
        return key in self._entries

    def __getitem__(self, key):
        return self._entries[key]

    def __setitem__(self, key, value):
        # NOTE: if Index is using a key object then we cannot check if key is valid. It's safer to use `add` method
        # instead of setting values directly.
        self._verify_and_get_key(object=value, key_object=None, key=key, missing_key_object_ok=True)

        self._entries[key] = value

    def __getstate__(self):
        return {
            "name": self._name,
            "object_type": get_type_name(self._object_type),
            "key_type": get_type_name(self._key_type),
            "attribute": self._attribute,
            "entries": self._entries,
        }

    def __setstate__(self, data):
        self._name = data.pop("name")
        self._object_type = get_class(data.pop("object_type"))
        self._key_type = get_class(data.pop("key_type"))
        self._attribute = data.pop("attribute")
        self._entries = data.pop("entries")

    def __iter__(self):
        return self._entries.__iter__()

    @property
    def name(self) -> str:
        """Return Index's name."""
        return self._name

    @property
    def object_type(self) -> type:
        """Return Index's object_type."""
        return self._object_type

    def get(self, key, default=None):
        """Return an entry based on its key.

        Args:
            key: The key of the entry to get.
            default: Default value to return of entry wasn't found (Default value = None).

        Returns:
            The found entry or the default value if it wasn't found.
        """
        return self._entries.get(key, default)

    def pop(self, key, default=MARKER):
        """Remove and return an object.

        Args:
            key: The key of the entry to remove.
            default: Default value to return of entry wasn't found (Default value = MARKER).

        Returns:
            The removed entry or the default value if it wasn't found.
        """
        if not key:
            return
        return self._entries.pop(key) if default is MARKER else self._entries.pop(key, default)

    def keys(self, min=None, max=None, excludemin=False, excludemax=False):
        """Return an iterator of keys."""
        return self._entries.keys(min=min, max=max, excludemin=excludemin, excludemax=excludemax)

    def values(self):
        """Return an iterator of values."""
        return self._entries.values()

    def items(self):
        """Return an iterator of keys and values."""
        return self._entries.items()

    def add(self, object: persistent.Persistent, *, key: Optional[str] = None, key_object=None, verify=True):
        """Update index with object.

        If `Index._attribute` is not None then key is automatically generated.
        Key is extracted from `key_object` if it is not None; otherwise, it's extracted from `object`.

        Args:
            object(persistent.Persistent): Object to add.
            key(Optional[str], optional): Key to use in the index (Default value = None).
            key_object: Object to use to extract a key from (Default value = None).
            verify: Whether to check if the key is valid (Default value = True).
        """
        assert isinstance(object, self._object_type), f"Cannot add objects of type '{type(object)}'"
        key = self._verify_and_get_key(
            object=object, key_object=key_object, key=key, missing_key_object_ok=False, verify=verify
        )
        self._entries[key] = object

    def remove(self, object: persistent.Persistent, *, key: Optional[str] = None, key_object=None, verify=True):
        """Remove object from the index.

        If `Index._attribute` is not None then key is automatically generated.
        Key is extracted from `key_object` if it is not None; otherwise, it's extracted from `object`.

        Args:
            object(persistent.Persistent): Object to add.
            key(Optional[str], optional): Key to use in the index (Default value = None).
            key_object: Object to use to extract a key from (Default value = None).
            verify: Whether to check if the key is valid (Default value = True).
        """
        assert isinstance(object, self._object_type), f"Cannot remove objects of type '{type(object)}'"
        key = self._verify_and_get_key(
            object=object, key_object=key_object, key=key, missing_key_object_ok=False, verify=verify
        )
        del self._entries[key]

    def generate_key(self, object: persistent.Persistent, *, key_object=None):
        """Return index key for an object.

        Key is extracted from `key_object` if it is not None; otherwise, it's extracted from `object`.

        Args:
            object(persistent.Persistent): The object to generate a key for.
            key_object: The object to derive a key from (Default value = None).

        Returns:
            A key for object.
        """
        return self._verify_and_get_key(object=object, key_object=key_object, key=None, missing_key_object_ok=False)

    def _verify_and_get_key(
        self, *, object: persistent.Persistent, key_object, key, missing_key_object_ok, verify=True
    ):
        if self._key_type:
            if not missing_key_object_ok:
                assert isinstance(key_object, self._key_type), f"Invalid key type: {type(key_object)} for '{self.name}'"
        else:
            assert key_object is None, f"Index '{self.name}' does not accept 'key_object'"

        if self._attribute:
            key_object = key_object or object
            correct_key = get_attribute(key_object, self._attribute)
            if key is not None:
                if verify:
                    assert key == correct_key, f"Incorrect key for index '{self.name}': '{key}' != '{correct_key}'"
                else:
                    correct_key = key
        else:
            assert key is not None, "No key is provided"
            correct_key = key

        return correct_key


class Storage:
    """Store Persistent objects on the disk."""

    OID_FILENAME_LENGTH = 64

    def __init__(self, path: Union[Path, str]):
        self.path = Path(path)
        self.zstd_compressor = zstd.ZstdCompressor()
        self.zstd_decompressor = zstd.ZstdDecompressor()

    def store(self, filename: str, data: Union[Dict, List], compress=False, absolute: bool = False):
        """Store object.

        Args:
            filename(str): Target file name to store data in.
            data(Union[Dict, List]): The data to store.
            compress(bool): Whether to compress the data or store it as plain json (Default value = False).
            absolute(bool): Whether filename is an absolute path (Default value = False).
        """
        assert isinstance(filename, str)

        if absolute:
            path = Path(filename)
        else:
            is_oid_path = len(filename) == Storage.OID_FILENAME_LENGTH
            if is_oid_path:
                path = self.path / filename[0:2] / filename[2:4] / filename
                path.parent.mkdir(parents=True, exist_ok=True)
            else:
                path = self.path / filename
                self.path.mkdir(parents=True, exist_ok=True)

        if compress:
            with open(path, "wb") as fb, self.zstd_compressor.stream_writer(fb) as compressor:
                with io.TextIOWrapper(compressor) as out:
                    json.dump(data, out, ensure_ascii=False)
        else:
            with open(path, "w") as ft:
                json.dump(data, ft, ensure_ascii=False, sort_keys=True, indent=2)

    def load(self, filename: str, absolute: bool = False):
        """Load data for object with object id oid.

        Args:
            filename(str): The file name of the data to load.
            absolute(bool): Whether the path is absolute or a filename inside the database (Default value: False).
        Returns:
            The loaded data in dictionary form.
        """
        assert isinstance(filename, str)

        if absolute:
            path = Path(filename)
        else:
            is_oid_path = len(filename) == Storage.OID_FILENAME_LENGTH
            if is_oid_path:
                path = self.path / filename[0:2] / filename[2:4] / filename
            else:
                path = self.path / filename

        if not path.exists():
            raise errors.ObjectNotFoundError(filename)

        with open(path, "rb") as file:
            header = int.from_bytes(file.read(4), "little")
            file.seek(0)
            if header == zstd.MAGIC_NUMBER:
                with self.zstd_decompressor.stream_reader(file) as zfile:
                    data = json.load(zfile)
            else:
                try:
                    data = json.load(file)
                except json.JSONDecodeError:
                    raise errors.MetadataCorruptError(path)
        return data


class ObjectWriter:
    """Serialize objects for storage in storage."""

    def __init__(self, database: Database):
        self._database: Database = database

    def serialize(self, object: persistent.Persistent):
        """Convert an object to JSON.

        Args:
            object(persistent.Persistent): Object to serialize.

        Returns:
            dict: Dictionary containing serialized data.
        """
        assert isinstance(object, persistent.Persistent), f"Cannot serialize object of type '{type(object)}': {object}"
        assert object._p_oid, f"Object does not have an oid: '{object}'"
        assert object._p_jar is not None, f"Object is not associated with a Database: '{object}'"

        self._serialization_cache: Dict[int, Any] = {}
        state = object.__getstate__()
        was_dict = isinstance(state, dict)
        data = self._serialize_helper(state)
        is_dict = isinstance(data, dict)

        if not is_dict or (is_dict and not was_dict):
            data = {"@renku_data_value": data}

        data["@renku_data_type"] = get_type_name(object)
        data["@renku_oid"] = object._p_oid

        return data

    def _serialize_helper(self, obj):
        # TODO: Raise an error if an unsupported object is being serialized
        if obj is None:
            return None
        elif isinstance(obj, (int, float, str, bool)):
            return obj
        elif isinstance(obj, list):
            return [self._serialize_helper(value) for value in obj]
        elif isinstance(obj, set):
            return {
                "@renku_data_type": SET_TYPE,
                "@renku_data_value": [self._serialize_helper(value) for value in obj],
            }
        elif isinstance(obj, frozenset):
            return {
                "@renku_data_type": FROZEN_SET_TYPE,
                "@renku_data_value": [self._serialize_helper(value) for value in obj],
            }
        elif isinstance(obj, dict):
            result = dict()
            items = sorted(obj.items(), key=lambda x: x[0])
            for key, value in items:
                result[key] = self._serialize_helper(value)
            return result
        elif isinstance(obj, Index):
            # NOTE: Index objects are not stored as references and are included in their parent object (i.e. root)
            state = obj.__getstate__()
            state = self._serialize_helper(state)
            return {"@renku_data_type": get_type_name(obj), "@renku_oid": obj._p_oid, **state}
        elif isinstance(obj, (OOTreeSet, Length, OOSet)):
            state = obj.__getstate__()
            state = self._serialize_helper(state)
            return {"@renku_data_type": get_type_name(obj), "@renku_data_value": state}
        elif isinstance(obj, persistent.Persistent):
            if not obj._p_oid:
                obj._p_oid = Database.generate_oid(obj)
            if obj._p_state not in [GHOST, UPTODATE] or (obj._p_state == UPTODATE and obj._p_serial == NEW):
                self._database.register(obj)
            return {"@renku_data_type": get_type_name(obj), "@renku_oid": obj._p_oid, "@renku_reference": True}
        elif isinstance(obj, datetime.datetime):
            value = obj.isoformat()
        elif isinstance(obj, tuple):
            value = tuple(self._serialize_helper(value) for value in obj)
        elif isinstance(obj, (InterfaceClass)):
            # NOTE: Zope interfaces are weird, they're a class with type InterfaceClass, but need to be deserialized
            # as the class (without instantiation)
            return {"@renku_data_type": TYPE_TYPE, "@renku_data_value": f"{obj.__module__}.{obj.__name__}"}
        elif isinstance(obj, type):
            # NOTE: We're storing a type, not an instance
            return {"@renku_data_type": TYPE_TYPE, "@renku_data_value": get_type_name(obj)}
        elif isinstance(obj, (FunctionType, BuiltinFunctionType)):
            name = obj.__name__
            module = getattr(obj, "__module__", None)
            return {"@renku_data_type": FUNCTION_TYPE, "@renku_data_value": f"{module}.{name}"}
        else:
            if id(obj) in self._serialization_cache:
                # NOTE: We already serialized this -> circular/repeat reference
                return {"@renku_data_type": REFERENCE_TYPE, "@renku_data_value": self._serialization_cache[id(obj)]}

            # NOTE: The reference used for circular reference is just the position in the serialization cache,
            # as the order is deterministic So the order in which objects are encoutered is their id for referencing.
            self._serialization_cache[id(obj)] = len(self._serialization_cache)
            if hasattr(obj, "__getstate__"):
                # NOTE: On Python 3.11+ this just returns __dict__ if __getstate__ isn't implemented.
                value = obj.__getstate__().copy()
            else:
                value = obj.__dict__.copy()
            value = {k: v for k, v in value.items() if not k.startswith("_v_")}
            value = self._serialize_helper(value)

        return {"@renku_data_type": get_type_name(obj), "@renku_data_value": value}


class ObjectReader:
    """Deserialize objects loaded from storage."""

    def __init__(self, database: Database):
        self._classes: Dict[str, type] = {}
        self._database = database

        # a cache for normal (non-persistent objects with an id) to deduplicate them on load
        self._normal_object_cache: Dict[str, Any] = {}
        self._deserialization_cache: List[Any] = []

    def _get_class(self, type_name: str) -> Optional[type]:
        cls = self._classes.get(type_name)
        if cls:
            return cls

        cls = get_class(type_name)

        if cls is None:
            return None

        self._classes[type_name] = cls
        return cls

    def set_ghost_state(self, object: persistent.Persistent, data: Dict):
        """Set state of a Persistent ghost object.

        Args:
            object(persistent.Persistent): Object to set state on.
            data(Dict): State to set on the object.
        """
        previous_cache = self._deserialization_cache
        self._deserialization_cache = []

        state = self._deserialize_helper(data, create=False)
        object.__setstate__(state)

        self._deserialization_cache = previous_cache

    def deserialize(self, data):
        """Convert JSON to Persistent object.

        Args:
            data: Data to deserialize.

        Returns:
            Deserialized object.
        """
        oid = data["@renku_oid"]

        self._deserialization_cache = []

        object = self._deserialize_helper(data)

        object._p_oid = oid
        object._p_jar = self._database

        return object

    def _deserialize_helper(self, data, create=True):
        if data is None:
            return None
        elif isinstance(data, (int, float, str, bool)):
            return data
        elif isinstance(data, list):
            return [self._deserialize_helper(value) for value in data]
        else:
            assert isinstance(data, dict), f"Data must be a dict: '{type(data)}'"

            if "@renku_data_type" not in data:  # NOTE: A normal dict value
                assert "@renku_oid" not in data
                items = sorted(data.items(), key=lambda x: x[0])
                for key, value in items:
                    data[key] = self._deserialize_helper(value)
                return data

            object_type = data.pop("@renku_data_type")
            if object_type in (TYPE_TYPE, FUNCTION_TYPE):
                # NOTE: if we stored a type (not instance), return the type
                return self._get_class(data["@renku_data_value"])
            elif object_type == REFERENCE_TYPE:
                # NOTE: we had a circular reference, we return the (not yet finalized) class here
                return self._deserialization_cache[data["@renku_data_value"]]
            elif object_type == SET_TYPE:
                return {self._deserialize_helper(value) for value in data["@renku_data_value"]}
            elif object_type == FROZEN_SET_TYPE:
                return frozenset([self._deserialize_helper(value) for value in data["@renku_data_value"]])

            cls = self._get_class(object_type)

            if cls is None:
                raise TypeError(f"Couldn't find class '{object_type}'")

            if issubclass(cls, datetime.datetime):
                assert create
                data = data["@renku_data_value"]
                return datetime.datetime.fromisoformat(data)
            elif issubclass(cls, tuple):
                data = data["@renku_data_value"]
                return tuple(self._deserialize_helper(value) for value in data)

            oid: str = data.pop("@renku_oid", None)
            if oid:
                assert isinstance(oid, str)

                if "@renku_reference" in data and data["@renku_reference"]:  # A reference
                    assert create, f"Cannot deserialize a reference without creating an instance {data}"
                    new_object = self._database.get_cached(oid)
                    if new_object is not None:
                        return new_object
                    assert issubclass(cls, persistent.Persistent)
                    new_object = cls.__new__(cls)
                    self._database.new_ghost(oid, new_object)
                    return new_object
                elif issubclass(cls, Index):
                    new_object = self._database.get_cached(oid)
                    if new_object:
                        return new_object
                    new_object = cls.__new__(cls)
                    new_object._p_oid = oid
                    self.set_ghost_state(new_object, data)
                    return new_object

            if "@renku_data_value" in data:
                data = data["@renku_data_value"]

            if not create:
                data = self._deserialize_helper(data)
                return data

            if issubclass(cls, persistent.Persistent):
                new_object = cls.__new__(cls)
                new_object._p_oid = oid  # type: ignore[attr-defined]
                self.set_ghost_state(new_object, data)
            elif issubclass(cls, Enum):
                # NOTE: Enum replaces __new__ on classes with its own versions that validates entries
                new_object = cls.__new__(cls, data["_value_"])
                return new_object
            else:
                new_object = cls.__new__(cls)

                # NOTE: we deserialize in the same order as we serialized, so the two stacks here match
                self._deserialization_cache.append(new_object)
                cache_index = len(self._deserialization_cache) - 1

                data = self._deserialize_helper(data)
                assert isinstance(data, dict)

                if "id" in data and data["id"] in self._normal_object_cache:
                    existing_object = self._normal_object_cache[data["id"]]

                    # NOTE: replace uninitialized object in cache with actual object
                    self._deserialization_cache[cache_index] = existing_object
                    return existing_object

                if hasattr(new_object, "__setstate__"):
                    new_object.__setstate__(data)
                else:
                    for name, value in data.items():
                        object.__setattr__(new_object, name, value)

                if issubclass(cls, Immutable):
                    new_object = cls.make_instance(new_object)

                if "id" in data and isinstance(data["id"], str) and data["id"].startswith("/"):
                    self._normal_object_cache[data["id"]] = new_object

            return new_object



#
# Copyright 2017-2023 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku plan gateway interface."""

from abc import ABC
from typing import Dict, List, Optional

from renku.domain_model.workflow.plan import AbstractPlan


class IPlanGateway(ABC):
    """Interface for the PlanGateway."""

    def get_by_id(self, id: Optional[str]) -> Optional[AbstractPlan]:
        """Get a plan by id."""
        raise NotImplementedError

    def get_by_name(self, name: str) -> Optional[AbstractPlan]:
        """Get a plan by name."""
        raise NotImplementedError

    def get_by_name_or_id(self, name_or_id: str) -> AbstractPlan:
        """Get a plan by name or id."""
        raise NotImplementedError()

    def list_by_name(self, starts_with: str, ends_with: Optional[str] = None) -> List[str]:
        """Search plans by name."""
        raise NotImplementedError

    def get_newest_plans_by_names(self, include_deleted: bool = False) -> Dict[str, AbstractPlan]:
        """Return a mapping of all plan names to their newest plans."""
        raise NotImplementedError

    def get_all_plans(self) -> List[AbstractPlan]:
        """Get all plans in project."""
        raise NotImplementedError

    def add(self, plan: AbstractPlan) -> None:
        """Add a plan to the database."""
        raise NotImplementedError



#
# Copyright 2018-2023 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Converter for workflows to Renku workflow description."""

import tempfile
from pathlib import Path
from typing import Any, Dict, Optional, Union

from renku.core import errors
from renku.core.plugin import hookimpl
from renku.core.util import communication
from renku.core.util.os import get_absolute_path
from renku.core.util.yaml import write_yaml
from renku.domain_model.workflow.composite_plan import CompositePlan
from renku.domain_model.workflow.converters import IWorkflowConverter
from renku.domain_model.workflow.plan import Plan


class RenkuWorkflowFileExporter(IWorkflowConverter):
    """Export a ``CompositePlan`` or a ``Plan`` to Renku workflow file."""

    @hookimpl
    def workflow_format(self):
        """Workflow format name."""
        return self, ["renku"]

    @hookimpl
    def workflow_convert(
        self,
        workflow: Union[CompositePlan, Plan],
        basedir: Path,
        output: Optional[Path],
        output_format: Optional[str],
        resolve_paths: Optional[bool],
        nest_workflows: Optional[bool],
    ) -> str:
        """Converts the specified workflow to renku workflow file."""

        def get_output_path():
            if output:
                path = output / workflow.name if output.is_dir() else output
            else:
                path = Path(tempfile.mkdtemp()) / workflow.name

            absolute_path = Path(get_absolute_path(path))
            if absolute_path.exists():
                communication.confirm(f"Output file '{path}' exists. Overwrite it?", abort=True, warning=True)

            return absolute_path

        output_path = get_output_path()

        if isinstance(workflow, CompositePlan):
            RenkuWorkflowFileExporter.convert_composite_plan(plan=workflow, path=output_path)
        else:
            RenkuWorkflowFileExporter.convert_plan(plan=workflow, path=output_path)

        return output_path.read_text()

    @staticmethod
    def convert_composite_plan(plan: CompositePlan, path: Path):
        """Converts a composite plan to a workflow file."""
        data: Dict[str, Any] = {
            "name": plan.name,
            "description": plan.description,
            "keywords": plan.keywords,
            "steps": [{p.name: RenkuWorkflowFileExporter.convert_step(step=p)} for p in plan.plans],
        }

        write_yaml(path=path, data=data)

    @staticmethod
    def convert_plan(plan: Plan, path: Path):
        """Converts a Plan to Renku workflow file format."""
        # NOTE: There is no root plan and we need to create one; its name can't be the same as step's (i.e. plan's) name
        data = {
            "name": f"{plan.name}-workflow-file",
            "steps": [{plan.name: RenkuWorkflowFileExporter.convert_step(step=plan)}],
        }
        write_yaml(path=path, data=data)

    @staticmethod
    def convert_step(step: Plan):
        """Converts a Plan to workflow file dict."""

        def convert_path(parameter):
            data = {"path": parameter.actual_value}
            if parameter.prefix:
                data["prefix"] = parameter.prefix.rstrip()
            if parameter.description:
                data["description"] = parameter.description

            return data

        def convert_parameter(parameter):
            data = {"value": parameter.actual_value}
            if parameter.prefix:
                data["prefix"] = parameter.prefix.rstrip()
            if parameter.description:
                data["description"] = parameter.description

            return data

        # NOTE: Check to make sure that the parent CompositePlan doesn't have a child CompositePlan
        if isinstance(step, CompositePlan):
            raise errors.WorkflowExportError(
                "Cannot export CompositePlans within a CompositePlan to Renku workflow file format"
            )

        step_data: Dict = {
            "command": " ".join(step.to_argv(with_streams=True)),
            "description": step.description,
            "keywords": step.keywords,
        }
        if step.inputs:
            step_data["inputs"] = [{p.name: convert_path(p)} for p in step.inputs]
        if step.outputs:
            step_data["outputs"] = [{p.name: convert_path(p)} for p in step.outputs]
        if step.parameters:
            step_data["parameters"] = [{p.name: convert_parameter(p)} for p in step.parameters]

        return step_data



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Logic for handling a data storage."""

import functools
import itertools
import os
import re
import shlex
import tempfile
from collections import defaultdict
from pathlib import Path
from shutil import move, which
from subprocess import PIPE, STDOUT, check_output, run
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Union

import pathspec

from renku.core import errors
from renku.core.config import get_value
from renku.core.constant import RENKU_LFS_IGNORE_PATH, RENKU_PROTECTED_PATHS
from renku.core.util import communication
from renku.core.util.git import get_in_submodules, run_command
from renku.core.util.os import expand_directories, parse_file_size
from renku.domain_model.project_context import project_context

if TYPE_CHECKING:
    from renku.infrastructure.repository import Repository


_CMD_STORAGE_INSTALL = ["git", "lfs", "install", "--local"]

_CMD_STORAGE_TRACK = ["git", "lfs", "track", "--"]

_CMD_STORAGE_UNTRACK = ["git", "lfs", "untrack", "--"]

_CMD_STORAGE_CLEAN = ["git", "lfs", "clean"]

_CMD_STORAGE_CHECKOUT = ["git", "lfs", "checkout"]

_CMD_STORAGE_PULL = ["git", "lfs", "pull", "-I"]

_CMD_STORAGE_MIGRATE_IMPORT = ["git", "lfs", "migrate", "import"]

_CMD_STORAGE_MIGRATE_INFO = ["git", "lfs", "migrate", "info", "--top", "42000"]

_CMD_STORAGE_LIST = ["git", "lfs", "ls-files", "-n"]

_CMD_STORAGE_STATUS = ["git", "lfs", "status"]

_LFS_HEADER = "version https://git-lfs.github.com/spec/"


class RenkuGitWildMatchPattern(pathspec.patterns.GitWildMatchPattern):
    """Custom GitWildMatchPattern matcher."""

    __slots__ = ("pattern",)

    def __init__(self, pattern, include=None):
        """Initialize RenkuRegexPattern."""
        super().__init__(pattern, include)
        self.pattern = pattern


pathspec.util.register_pattern("renku_gitwildmatch", RenkuGitWildMatchPattern)


def check_external_storage_wrapper(fn):
    """Check availability of external storage on methods that need it.

    Raises:
        ``errors.ExternalStorageNotInstalled``: If external storage isn't installed.
        ``errors.ExternalStorageDisabled``: If external storage isn't enabled.
    """

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        if not check_external_storage():
            pass
        else:
            return fn(*args, **kwargs)

    return wrapper


@functools.lru_cache
def storage_installed() -> bool:
    """Verify that git-lfs is installed and on system PATH."""
    return bool(which("git-lfs"))


def storage_installed_locally() -> bool:
    """Verify that git-lfs is installed for the project."""
    repo_config = project_context.repository.get_configuration(scope="local")
    return repo_config.has_section('filter "lfs"')


def check_external_storage():
    """Check if repository has external storage enabled.

    Raises:
        ``errors.ExternalStorageNotInstalled``: If external storage isn't installed.
        ``errors.ExternalStorageDisabled``: If external storage isn't enabled.
    """
    installed_locally = storage_installed_locally()
    is_storage_installed = installed_locally and storage_installed()

    if project_context.external_storage_requested and not is_storage_installed:
        raise errors.ExternalStorageDisabled()

    if installed_locally and not storage_installed():
        raise errors.ExternalStorageNotInstalled()

    return is_storage_installed


def renku_lfs_ignore() -> pathspec.PathSpec:
    """Gets pathspec for files to not add to LFS."""
    ignore_path = project_context.path / RENKU_LFS_IGNORE_PATH

    if not os.path.exists(ignore_path):
        return pathspec.PathSpec.from_lines("renku_gitwildmatch", RENKU_PROTECTED_PATHS)
    with ignore_path.open("r") as f:
        # NOTE: Append `renku_protected_paths` at the end to give it the highest priority
        lines = itertools.chain(f, RENKU_PROTECTED_PATHS)
        return pathspec.PathSpec.from_lines("renku_gitwildmatch", lines)


def get_minimum_lfs_file_size() -> int:
    """The minimum size of a file in bytes to be added to lfs."""
    size = get_value("renku", "lfs_threshold")

    return parse_file_size(size)


def init_external_storage(force: bool = False) -> None:
    """Initialize the external storage for data."""
    try:
        result = run(
            _CMD_STORAGE_INSTALL + (["--force"] if force else []),
            stdout=PIPE,
            stderr=STDOUT,
            cwd=project_context.path,
            text=True,
        )

        if result.returncode != 0:
            raise errors.GitLFSError(f"Error executing 'git lfs install: \n {result.stdout}")
    except (KeyboardInterrupt, OSError) as e:
        raise errors.ParameterError(f"Couldn't run 'git lfs':\n{e}")


@check_external_storage_wrapper
def track_paths_in_storage(*paths: Union[Path, str]) -> Optional[List[str]]:
    """Track paths in the external storage."""
    if not project_context.external_storage_requested or not check_external_storage():
        return None

    # Calculate which paths can be tracked in lfs
    track_paths: List[str] = []
    attrs = project_context.repository.get_attributes(*paths)

    for path in paths:
        path = Path(path)

        # Do not track symlinks in LFS
        if path.is_symlink():
            continue

        # Do not add files with filter=lfs in .gitattributes
        if attrs.get(str(path), {}).get("filter") == "lfs" or not (project_context.path / path).exists():
            continue

        relative_path = Path(path).relative_to(project_context.path) if path.is_absolute() else path

        if (
            path.is_dir()
            and not renku_lfs_ignore().match_file(relative_path)
            and not any(renku_lfs_ignore().match_tree(str(relative_path)))
        ):
            track_paths.append(str(path / "**"))
        elif not renku_lfs_ignore().match_file(str(relative_path)):
            file_size = os.path.getsize(str(os.path.relpath(project_context.path / path, os.getcwd())))
            if file_size >= get_minimum_lfs_file_size():
                track_paths.append(str(relative_path))

    if track_paths:
        try:
            result = run_command(
                _CMD_STORAGE_TRACK,
                *track_paths,
                stdout=PIPE,
                stderr=STDOUT,
                cwd=project_context.path,
                universal_newlines=True,
            )

            if result and result.returncode != 0:
                raise errors.GitLFSError(f"Error executing 'git lfs track: \n {result.stdout}")
        except (KeyboardInterrupt, OSError) as e:
            raise errors.ParameterError(f"Couldn't run 'git lfs':\n{e}")

    show_message = get_value("renku", "show_lfs_message")
    if track_paths and (show_message is None or show_message.lower() == "true"):
        files_list = "\n\t".join(track_paths)
        communication.info(
            f"Adding these files to Git LFS:\n\t{files_list}"
            "\nTo disable this message in the future, run:\n\trenku config set show_lfs_message false"
        )

    return track_paths


@check_external_storage_wrapper
def untrack_paths_from_storage(*paths: Union[Path, str]) -> None:
    """Untrack paths from the external storage."""
    try:
        result = run_command(
            _CMD_STORAGE_UNTRACK,
            *paths,
            stdout=PIPE,
            stderr=STDOUT,
            cwd=project_context.path,
            universal_newlines=True,
        )

        if result and result.returncode != 0:
            raise errors.GitLFSError(f"Error executing 'git lfs untrack: \n {result.stdout}")
    except (KeyboardInterrupt, OSError) as e:
        raise errors.ParameterError(f"Couldn't run 'git lfs':\n{e}")


@check_external_storage_wrapper
def list_tracked_paths() -> List[Path]:
    """List paths tracked in lfs."""
    try:
        files = check_output(_CMD_STORAGE_LIST, cwd=project_context.path, encoding="UTF-8")
    except (KeyboardInterrupt, OSError) as e:
        raise errors.ParameterError(f"Couldn't run 'git lfs ls-files':\n{e}")
    files_split: List[Path] = [project_context.path / f for f in files.splitlines()]
    return files_split


@check_external_storage_wrapper
def list_unpushed_lfs_paths(repository: "Repository") -> List[Path]:
    """List paths tracked in lfs for a repository."""

    if len(repository.remotes) < 1 or (repository.active_branch and not repository.active_branch.remote_branch):
        raise errors.GitConfigurationError(
            f"No git remote is configured for {project_context.path} branch "
            + f"{repository.active_branch.name}."  # type: ignore
            + "Cleaning the storage cache would lead to a loss of data as "
            + "it is not on a server. Please see "
            + "https://www.atlassian.com/git/tutorials/syncing for "
            + "information on how to sync with a remote."
        )
    try:
        status = check_output(_CMD_STORAGE_STATUS, cwd=project_context.path, encoding="UTF-8")
    except (KeyboardInterrupt, OSError) as e:
        raise errors.ParameterError(f"Couldn't run 'git lfs status':\n{e}")

    files = status.split("Objects to be committed:")[0].splitlines()[2:]
    return [project_context.path / f.rsplit("(", 1)[0].strip() for f in files if f.strip()]


@check_external_storage_wrapper
def pull_paths_from_storage(repository: "Repository", *paths: Union[Path, str]):
    """Pull paths from LFS."""
    project_dict = defaultdict(list)

    for path in expand_directories(paths):
        sub_repository, _, path = get_in_submodules(repository, repository.head.commit, path)
        try:
            absolute_path = Path(path).resolve()
            relative_path = absolute_path.relative_to(project_context.path)
        except ValueError:  # An external file
            continue

        project_dict[sub_repository.path].append(shlex.quote(str(relative_path)))

    for project_path, file_paths in project_dict.items():
        result = run_command(
            _CMD_STORAGE_PULL,
            *file_paths,
            separator=",",
            cwd=project_path,
            stdout=PIPE,
            stderr=STDOUT,
            universal_newlines=True,
        )

        if result and result.returncode != 0:
            raise errors.GitLFSError(f"Cannot pull LFS objects from server:\n {result.stdout}")


@check_external_storage_wrapper
def clean_storage_cache(*check_paths: Union[Path, str]) -> Tuple[List[str], List[str]]:
    """Remove paths from lfs cache."""
    project_dict = defaultdict(list)
    repositories: Dict[Path, "Repository"] = {}
    tracked_paths: Dict[Path, List[Path]] = {}
    unpushed_paths: Dict[Path, List[Path]] = {}
    untracked_paths: List[str] = []
    local_only_paths: List[str] = []

    repository = project_context.repository

    for path in expand_directories(check_paths):
        current_repository, _, path = get_in_submodules(repository=repository, commit=repository.head.commit, path=path)
        try:
            absolute_path = Path(path).resolve()
            relative_path = absolute_path.relative_to(project_context.path)
        except ValueError:  # An external file
            continue

        if project_context.path not in tracked_paths:
            tracked_paths[project_context.path] = list_tracked_paths()

        if project_context.path not in unpushed_paths:
            u_paths = list_unpushed_lfs_paths(current_repository)
            unpushed_paths[project_context.path] = u_paths

        if absolute_path in unpushed_paths[project_context.path]:
            local_only_paths.append(str(relative_path))
        elif absolute_path not in tracked_paths[project_context.path]:
            untracked_paths.append(str(relative_path))
        else:
            project_dict[project_context.path].append(str(relative_path))
            repositories[project_context.path] = current_repository

    for project_path, paths in project_dict.items():
        current_repository = repositories[project_path]

        for path in paths:
            with open(path) as tracked_file:
                try:
                    header = tracked_file.read(len(_LFS_HEADER))
                    if header == _LFS_HEADER:
                        # file is not pulled
                        continue
                except UnicodeDecodeError:
                    # likely a binary file, not lfs pointer file
                    pass
            with tempfile.NamedTemporaryFile(mode="w+t", encoding="utf-8", delete=False) as tmp, open(
                path, "r+t"
            ) as input_file:
                result = run(_CMD_STORAGE_CLEAN, cwd=project_path, stdin=input_file, stdout=tmp, text=True)

                if result.returncode != 0:
                    raise errors.GitLFSError(f"Error executing 'git lfs clean: \n {result.stdout}")

                tmp_path = tmp.name
            move(tmp_path, path)

            # get lfs sha hash
            old_pointer = current_repository.get_raw_content(path=path, revision="HEAD")
            old_pointer = old_pointer.splitlines()[1]
            old_pointer = old_pointer.split(" ")[1].split(":")[1]

            prefix1 = old_pointer[:2]
            prefix2 = old_pointer[2:4]

            # remove from lfs cache
            object_path = project_context.path / ".git" / "lfs" / "objects" / prefix1 / prefix2 / old_pointer
            object_path.unlink()

        # add paths so they don't show as modified
        current_repository.add(*paths)

    return untracked_paths, local_only_paths


@check_external_storage_wrapper
def checkout_paths_from_storage(*paths: Union[Path, str]):
    """Checkout a paths from LFS."""
    result = run_command(
        _CMD_STORAGE_CHECKOUT,
        *paths,
        cwd=project_context.path,
        stdout=PIPE,
        stderr=STDOUT,
        universal_newlines=True,
    )

    if result and result.returncode != 0:
        raise errors.GitLFSError(f"Error executing 'git lfs checkout: \n {result.stdout}")


def check_requires_tracking(*paths: Union[Path, str]) -> Optional[List[str]]:
    """Check paths and return a list of those that must be tracked."""

    if not project_context.external_storage_requested:
        return None

    attrs = project_context.repository.get_attributes(*paths)
    track_paths: List[str] = []

    for path in paths:
        absolute_path = Path(os.path.abspath(project_context.path / path))
        path = str(path)

        # Do not track symlinks in LFS
        if absolute_path.is_symlink():
            continue

        # Do not add files with filter=lfs in .gitattributes
        if attrs.get(path, {}).get("filter") == "lfs":
            continue

        if not absolute_path.is_dir():
            if renku_lfs_ignore().match_file(path):
                continue
            if os.path.getsize(absolute_path) < get_minimum_lfs_file_size():
                continue

            track_paths.append(path)

    return track_paths


def get_lfs_migrate_filters() -> Tuple[List[str], List[str]]:
    """Gets include, exclude and above filters for lfs migrate."""

    def add_migrate_pattern(pattern, collection):
        if pattern in RENKU_PROTECTED_PATHS:
            return
        pattern = pattern.strip()
        if pattern.endswith("*"):
            return
        pattern = pattern.rstrip("/")
        collection.append(f"{pattern}/**")

    includes = []
    excludes = []
    for p in renku_lfs_ignore().patterns:
        if p.regex is None:
            continue

        pattern = p.pattern.replace(os.linesep, "").replace("\n", "")
        if pattern.startswith("!"):
            pattern = pattern.replace("!", "", 1)

        if p.include:  # File ignored by LFS
            excludes.append(pattern)
            add_migrate_pattern(pattern, excludes)
        else:
            includes.append(pattern)
            add_migrate_pattern(pattern, includes)

    if excludes:
        excludes = ["--exclude", ",".join(excludes)]
    if includes:
        includes = ["--include", ",".join(includes)]

    return includes, excludes


def check_lfs_migrate_info(everything: bool = False, use_size_filter: bool = True) -> List[str]:
    """Return list of file groups in history should be in LFS."""
    ref = (
        ["--everything"]
        if everything or not project_context.repository.active_branch
        else ["--include-ref", project_context.repository.active_branch.name]
    )

    includes, excludes = get_lfs_migrate_filters()

    ignore_pointers = ["--pointers", "ignore"]

    command = _CMD_STORAGE_MIGRATE_INFO + ref + includes + excludes

    # NOTE: ``lfs migrate info`` supports ``--above`` while ``lfs migrate import`` doesn't.
    if use_size_filter:
        above = ["--above", str(get_minimum_lfs_file_size())]
        command += above

    try:
        lfs_output = run(
            command + ignore_pointers,
            stdout=PIPE,
            stderr=STDOUT,
            cwd=project_context.path,
            text=True,
        )
    except (KeyboardInterrupt, OSError) as e:
        raise errors.GitError(f"Couldn't run 'git lfs migrate info':\n{e}")

    if lfs_output.returncode != 0:
        # NOTE: try running without --pointers (old versions of git lfs)
        try:
            lfs_output = run(command, stdout=PIPE, stderr=STDOUT, cwd=project_context.path, text=True)
        except (KeyboardInterrupt, OSError) as e:
            raise errors.GitError(f"Couldn't run 'git lfs migrate info':\n{e}")

        if lfs_output.returncode != 0:
            raise errors.GitLFSError(f"Error executing 'git lfs migrate info: \n {lfs_output.stdout}")

    groups: List[str] = []
    files_re = re.compile(r"(.*\s+[\d.]+\s+\S+).*")

    for line in lfs_output.stdout.split("\n"):
        match = files_re.match(line)
        if match:
            groups.append(match.groups()[0])

    if groups and use_size_filter:
        # NOTE: Since there are some large files, remove the size filter so that users get list of all files that
        # will be moved to LFS.
        return check_lfs_migrate_info(everything=everything, use_size_filter=False)

    return groups


def migrate_files_to_lfs(paths: List[str]):
    """Migrate files to Git LFS."""
    if paths:
        includes: List[str] = ["--include", ",".join(paths)]
        excludes: List[str] = []
    else:
        includes, excludes = get_lfs_migrate_filters()

    command = _CMD_STORAGE_MIGRATE_IMPORT + includes + excludes

    try:
        lfs_output = run(command, stdout=PIPE, stderr=STDOUT, cwd=project_context.path, text=True)
    except (KeyboardInterrupt, OSError) as e:
        raise errors.GitError(f"Couldn't run 'git lfs migrate import':\n{e}")

    if lfs_output.returncode != 0:
        raise errors.GitLFSError(f"Error executing 'git lfs migrate import: \n {lfs_output.stdout}")



#
# Copyright 2018-2023 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Manage an external storage.

Commands and options
~~~~~~~~~~~~~~~~~~~~

.. rst-class:: cli-reference-commands

.. click:: renku.ui.cli.storage:storage
   :prog: renku storage
   :nested: full

Pulling files from git LFS
~~~~~~~~~~~~~~~~~~~~~~~~~~

LFS works by checking small pointer files into git and saving the actual
contents of a file in LFS. If instead of your file content, you see
something like this, it means the file is stored in git LFS and its
contents are not currently available locally (they are not pulled):

.. code-block:: console

    version https://git-lfs.github.com/spec/v1
    oid sha256:42b5c7fb2acd54f6d3cd930f18fee3bdcb20598764ca93bdfb38d7989c054bcf
    size 12

You can manually pull contents of file(s) you want with:

.. code-block:: console

    $ renku storage pull file1 file2

.. cheatsheet::
   :group: Misc
   :command: $ renku storage pull <path>...
   :description: Pull <path>'s from external storage (LFS).
   :target: rp

Removing local content of files stored in git LFS
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

If you want to restore a file back to its pointer file state, for instance
to free up space locally, you can run:

.. code-block:: console

    $ renku storage clean file1 file2

This removes any data cached locally for files tracked in in git LFS.

Migrate large files to git LFS
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

If you accidentally checked a large file into git or are moving a non-LFS
renku repo to git LFS, you can use the following command to migrate the files
to LFS:

.. code-block:: console

    $ renku storage migrate --all

This will move all files that are not excluded by `.renkulfsignore` into git
LFS.

.. note::

    Recent versions of Git LFS don't support filtering files based on their
    size. Therefore, Renku ignores `lfs_threshold` config value when migrating
    files to LFS using this command.

To only migrate specific files, you can also pass their paths to the command
like:

.. code-block:: console

    $ renku storage migrate big_file other_big_file
"""
import os

import click

import renku.ui.cli.utils.color as color
from renku.command.util import WARNING
from renku.ui.cli.utils.callback import ClickCallback


@click.group()
def storage():
    """Manage an external storage."""


@storage.command()
@click.argument("paths", type=click.Path(exists=True, dir_okay=True), nargs=-1, required=True)
def pull(paths):
    """Pull the specified paths from external storage."""
    from renku.command.storage import pull_command

    pull_command().build().execute(paths=paths)


@storage.command()
@click.argument("paths", type=click.Path(exists=True, dir_okay=True), nargs=-1, required=True)
def clean(paths):
    """Remove files from lfs cache/turn them back into pointer files."""
    from renku.command.storage import clean_command

    communicator = ClickCallback()
    clean_command().with_communicator(communicator).build().execute(paths=paths)

    click.secho("OK", fg=color.GREEN)


@storage.command("check-lfs-hook", hidden=True)
@click.argument("paths", type=click.Path(exists=True, dir_okay=True), nargs=-1, required=True)
def check_lfs_hook(paths):
    """Check specified paths are tracked in external storage."""
    from renku.command.storage import check_lfs_hook_command

    paths = check_lfs_hook_command().build().execute(paths=paths).output
    if paths:
        click.echo(os.linesep.join(paths))
        exit(1)


@storage.command()
@click.option("--all", is_flag=True, help="Include all branches.")
def check(all):
    """Check if large files are committed to Git history."""
    from renku.command.storage import check_lfs_command

    files = check_lfs_command().build().execute(everything=all).output
    if files:
        message = WARNING + "Git history contains large files\n\t" + "\n\t".join(files)
        click.echo(message)
        exit(1)
    else:
        click.secho("OK", fg=color.GREEN)


@storage.command()
@click.option("--all", "-a", "migrate_all", is_flag=True, default=False, help="Migrate all large files not in git LFS.")
@click.argument("paths", type=click.Path(exists=True, dir_okay=True), nargs=-1)
def migrate(migrate_all, paths):
    """Migrate large files committed to git by moving them to LFS."""
    from renku.command.storage import check_lfs_command, fix_lfs_command

    if not paths:
        if not migrate_all:
            click.echo("Please specify paths to migrate or use the --all flag to migrate all large files.")
            exit(1)

        lfs_paths = check_lfs_command().build().execute(everything=migrate_all).output

        if not lfs_paths:
            click.echo("All files are already in LFS")
            exit(0)

        if not click.confirm("The following files will be moved to Git LFS:\n\t" + "\n\t".join(lfs_paths)):
            exit(0)

    fix_lfs_command().build().execute(paths)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper utilities for handling URLs."""

import os
import re
import unicodedata
import urllib
from pathlib import Path
from typing import List, Optional, Tuple
from urllib.parse import ParseResult, urlparse

from renku.core import errors
from renku.core.config import get_value
from renku.core.util.git import get_remote, parse_git_url
from renku.core.util.os import is_subpath
from renku.domain_model.project_context import project_context


def url_to_string(url):
    """Convert url from ``list`` or ``ParseResult`` to string."""
    if isinstance(url, list):
        return ParseResult(scheme=url[0], netloc=url[1], path=url[2], params="", query="", fragment="").geturl()

    if isinstance(url, ParseResult):
        return url.geturl()

    if isinstance(url, str):
        return url

    raise ValueError("url value not recognized")


def remove_credentials(url):
    """Remove username and password from a URL."""
    if url is None:
        return ""
    parsed = urllib.parse.urlparse(url)
    return parsed._replace(netloc=parsed.hostname).geturl()


def get_host(use_project_context: bool = True):
    """Return the hostname for the resource URIs.

    Default is localhost. If RENKU_DOMAIN is set, it overrides the host from remote.
    """
    host = "localhost"

    if use_project_context:
        host = project_context.remote.host or host

    return os.environ.get("RENKU_DOMAIN") or host


def get_path(url: str) -> str:
    """Return path part of a url."""
    return urllib.parse.urlparse(url).path


def get_scheme(uri: str) -> str:
    """Return scheme of a URI."""
    return urllib.parse.urlparse(uri).scheme.lower()


def parse_authentication_endpoint(endpoint: Optional[str] = None, use_remote: bool = False):
    """Return a parsed url.

    If an endpoint is provided then use it, otherwise, look for a configured endpoint. If no configured endpoint exists
    then try to use project's remote url.
    """
    if not endpoint:
        endpoint = get_value(section="renku", key="endpoint")
        if not endpoint:
            if not use_remote:
                return
            remote = get_remote(project_context.repository)
            if not remote or not remote.url:
                return
            endpoint = f"https://{parse_git_url(remote.url).hostname}/"

    if not endpoint.startswith("http"):
        endpoint = f"https://{endpoint}"

    parsed_endpoint = urllib.parse.urlparse(endpoint)
    if not parsed_endpoint.netloc:
        raise errors.ParameterError(f"Invalid endpoint: `{endpoint}`.")

    return parsed_endpoint._replace(scheme="https", path="/", params="", query="", fragment="")


def get_slug(name: str, invalid_chars: Optional[List[str]] = None, lowercase: bool = True) -> str:
    """Create a slug from name."""
    invalid_chars = invalid_chars or []
    lower_case = name.lower() if lowercase else name
    no_space = re.sub(r"\s+", "_", lower_case)
    normalized = unicodedata.normalize("NFKD", no_space).encode("ascii", "ignore").decode("utf-8")

    valid_chars_pattern = [r"\w", ".", "_", "-"]
    if len(invalid_chars) > 0:
        valid_chars_pattern = [ch for ch in valid_chars_pattern if ch not in invalid_chars]

    no_invalid_characters = re.sub(f'[^{"".join(valid_chars_pattern)}]', "_", normalized)
    no_duplicates = re.sub(r"([._-])[._-]+", r"\1", no_invalid_characters)
    valid_start = re.sub(r"^[._-]", "", no_duplicates)
    valid_end = re.sub(r"[._-]$", "", valid_start)
    no_dot_lock_at_end = re.sub(r"\.lock$", "_lock", valid_end)
    return no_dot_lock_at_end


def is_uri_subfolder(uri: str, subfolder_uri: str) -> bool:
    """Check if one uri is a 'subfolder' of another."""
    parsed_uri = urlparse(uri)
    parsed_subfolder_uri = urlparse(subfolder_uri)
    parsed_uri_path = parsed_uri.path
    parsed_subfolder_uri_path = parsed_subfolder_uri.path
    if parsed_uri_path in ["", "."]:
        # NOTE: s3://test has a path that equals "" and Path("") gets interpreted as Path(".")
        # this becomes a problem then when s3://test/1 has an "absolute-like" path of Path("/1")
        # and Path(".") is not considered a subpath of Path("/1") but from the uris we see that this
        # is indeed a subpath
        parsed_uri_path = "/"
    if parsed_subfolder_uri_path in ["", "."]:
        parsed_subfolder_uri_path = "/"
    if parsed_uri.scheme != parsed_subfolder_uri.scheme:
        # INFO: catch s3://test vs http://test
        return False
    if parsed_uri.netloc != parsed_subfolder_uri.netloc:
        # INFO: catch s3://test1 vs s3://test2
        return False
    return is_subpath(parsed_subfolder_uri_path, parsed_uri_path)


def resolve_uri(uri: str) -> str:
    """Resolve path part of a URI if it's a local URI."""
    scheme = get_scheme(uri)

    if scheme not in ("file", ""):
        return uri

    path = get_path(uri)
    path = str(Path(path).resolve())

    return f"{scheme}://{path}" if scheme else path


def check_url(url: str) -> Tuple[bool, bool]:
    """Check if a url is local/remote and if it contains a git repository."""
    # NOTE: Supported scheme before refactoring were: "", "file", "http", "https", "git+https", "git+ssh"
    u = urllib.parse.urlparse(url)
    scheme = u.scheme.lower()
    starts_with_git = url.lower().startswith("git@")

    is_remote = scheme not in ("", "file") or starts_with_git
    is_git = is_remote and (u.path.lower().endswith(".git") or scheme in ("git+https", "git+ssh") or starts_with_git)

    return is_remote, is_git



#
# Copyright 2018-2023 Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku cli for history of renku commands.

Description
~~~~~~~~~~~

Get the history of renku commands.
At the moment, it shows workflow executions and dataset changes.

Commands and options
~~~~~~~~~~~~~~~~~~~~

.. rst-class:: cli-reference-commands

.. click:: renku.ui.cli.log:log
   :prog: renku log
   :nested: full

Examples
~~~~~~~~~~~

.. code-block:: console

    $ renku log
    Activity /activities/be60896d8d984a0bb585e53f7a3146dc
    Start Time: 2022-02-03T13:56:27+01:00
    End Time: 2022-02-03T13:56:28+01:00
    User: John Doe <John.Doe@example.com>
    Renku Version: renku 1.0.5
    Plan:
        Id: /plans/c00826a571a246e79b0a3d77712e6f3b
        Name: python-test-fc2ec
    Command: python test.py
    Inputs:
            input-1: test.py
    Parameters:
            text: hello

    Dataset testset
    Date: 2022-02-03T11:26:55+01:00
    Changes: created
    Title set to: testset
    Creators modified:
            + John Doe <John.Doe@example.com>

To show only dataset entries, use ``-d``, to show only workflows, use ``-w``.

You can select a format using the ``--format <format>`` argument.

.. cheatsheet::
   :group: Misc
   :command: $ renku log
   :description: Show a history of renku actions.
   :target: rp
"""

import click

import renku.ui.cli.utils.color as color
from renku.command.view_model.log import (
    LOG_COLUMNS,
    LOG_FORMATS,
    ActivityLogViewModel,
    DatasetLogViewModel,
    LogViewModel,
)
from renku.ui.cli.utils.terminal import show_text_with_pager, strip_ansi_codes


def _print_log(log_entry: LogViewModel) -> str:
    """Turn a log entry into a printable string."""
    if isinstance(log_entry, ActivityLogViewModel):
        return _print_activity_log(log_entry)
    elif isinstance(log_entry, DatasetLogViewModel):
        return _print_dataset_log(log_entry)

    raise NotImplementedError()


def _print_activity_log(log_entry: ActivityLogViewModel) -> str:
    """Turn an activity log entry into a printable string."""
    from renku.ui.cli.utils.terminal import style_header, style_key

    results = [
        style_header(f"Activity {log_entry.id}"),
        style_key("Start Time: ") + log_entry.details.start_time,
        style_key("End Time: ") + log_entry.details.end_time,
    ]

    if log_entry.details.user:
        results.append(style_key("User: ") + log_entry.details.user)
    if log_entry.details.renku_version:
        results.append(style_key("Renku Version: ") + log_entry.details.renku_version)

    results.append(style_key("Plan:"))
    results.append(style_key("\tId: ") + log_entry.plan.id)
    results.append(style_key("\tName: ") + log_entry.plan.name)

    results.append(style_key("Command: ") + log_entry.description)

    if log_entry.details.inputs:
        results.append(style_key("Inputs:\n\t") + "\n\t".join(f"{i[0]}: {i[1]}" for i in log_entry.details.inputs))
    if log_entry.details.outputs:
        results.append(style_key("Outputs:\n\t") + "\n\t".join(f"{o[0]}: {o[1]}" for o in log_entry.details.outputs))
    if log_entry.details.parameters:
        results.append(
            style_key("Parameters:\n\t") + "\n\t".join(f"{p[0]}: {p[1]}" for p in log_entry.details.parameters)
        )

    return "\n".join(results)


def _print_dataset_log(log_entry: DatasetLogViewModel) -> str:
    """Turn a dataset log entry into a printable string."""
    from renku.ui.cli.utils.terminal import style_header, style_key

    results = [style_header(f"Dataset {log_entry.id}"), style_key("Date: ") + log_entry.date.isoformat()]
    change = []
    if log_entry.details.created:
        change.append("created")
    if log_entry.details.imported:
        change.append("imported")
    if log_entry.details.deleted:
        change.append("deleted")
    if log_entry.details.migrated:
        change.append("migrated")
    if log_entry.details.modified:
        change.append("modified")

    if change:
        results.append(style_key("Changes: ") + ", ".join(change))

    if log_entry.details.source:
        results.append(style_key("Source: ") + log_entry.details.source)

    if log_entry.details.title_changed:
        results.append(style_key("Title set to: ") + log_entry.details.title_changed)

    if log_entry.details.description_changed:
        results.append(style_key("Description set to: ") + log_entry.details.description_changed)

    if log_entry.details.files_added or log_entry.details.files_removed:
        added = ""
        removed = ""

        if log_entry.details.files_added:
            added = click.style("\n\t".join(f"+ {f}" for f in log_entry.details.files_added), fg=color.GREEN)

        if log_entry.details.files_removed:
            removed = ("\n\t" if added else "") + click.style(
                "\n\t".join(f"- {f}" for f in log_entry.details.files_removed), fg=color.RED
            )

        results.append(style_key("Files modified:\n\t") + added + removed)

    if log_entry.details.creators_added or log_entry.details.creators_removed:
        added = ""
        removed = ""

        if log_entry.details.creators_added:
            added = click.style("\n\t".join(f"+ {c}" for c in log_entry.details.creators_added), fg=color.GREEN)

        if log_entry.details.creators_removed:
            removed = ("\n\t" if added else "") + click.style(
                "\n\t".join(f"- {c}" for c in log_entry.details.creators_removed), fg=color.RED
            )

        results.append(style_key("Creators modified:\n\t") + added + removed)

    if log_entry.details.keywords_added or log_entry.details.keywords_removed:
        added = ""
        removed = ""

        if log_entry.details.keywords_added:
            added = click.style("\n\t".join(f"+ {k}" for k in log_entry.details.keywords_added), fg=color.GREEN)

        if log_entry.details.keywords_removed:
            removed = ("\n\t" if added else "") + click.style(
                "\n\t".join(f"- {k}" for k in log_entry.details.keywords_removed), fg=color.RED
            )

        results.append(style_key("Keywords modified:\n\t") + added + removed)

    if log_entry.details.images_changed_to:
        results.append(
            style_key("Images set to:\n\t")
            + click.style("\n\t".join(f"{i}" for i in log_entry.details.images_changed_to), fg=color.GREEN)
        )

    return "\n".join(results)


@click.command()
@click.option(
    "-c",
    "--columns",
    type=click.STRING,
    default="date,type,description",
    metavar="<columns>",
    help="Comma-separated list of column to display (for 'tabular' format): {}.".format(", ".join(LOG_COLUMNS.keys())),
    show_default=True,
)
@click.option(
    "--format",
    type=click.Choice(list(LOG_FORMATS.keys())),
    default="detailed",
    help="Choose an output format (default: detailed).",
)
@click.option("-w", "--workflows", is_flag=True, default=False, help="Show only workflow executions.")
@click.option("-d", "--datasets", is_flag=True, default=False, help="Show only dataset modifications.")
@click.option("--no-pager", is_flag=True, help="Don't use pager (less) for output.")
@click.option("-c", "--no-color", is_flag=True, help="Do not colorize output.")
def log(columns, format, workflows, datasets, no_pager, no_color):
    """Show a history of renku workflow and dataset commands."""
    from renku.command.log import log_command

    result = log_command().with_database().build().execute(workflows_only=workflows, datasets_only=datasets).output
    if format == "detailed":
        entries = sorted(result, key=lambda e: e.date, reverse=True)
        text = "\n\n".join([_print_log(e) for e in entries])

        if no_color:
            text = strip_ansi_codes(text)
    else:
        text = LOG_FORMATS[format](result, columns)  # type: ignore

    if no_pager:
        click.echo(text)
    else:
        show_text_with_pager(text)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Models representing datasets."""

import copy
import os
import posixpath
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Optional, Union, cast
from urllib.parse import quote, urlparse
from uuid import uuid4

import deal
import marshmallow

from renku.core import errors
from renku.core.util.datetime8601 import fix_datetime, local_now, parse_date
from renku.core.util.git import get_entity_from_revision
from renku.core.util.metadata import is_linked_file
from renku.core.util.os import get_absolute_path
from renku.core.util.urls import get_path, get_slug
from renku.domain_model.constant import NO_VALUE, NON_EXISTING_ENTITY_CHECKSUM
from renku.domain_model.project_context import project_context
from renku.infrastructure.immutable import Immutable, Slots
from renku.infrastructure.persistent import Persistent

if TYPE_CHECKING:
    from renku.domain_model.entity import Entity
    from renku.domain_model.provenance.agent import Person
    from renku.domain_model.provenance.annotation import Annotation


def is_dataset_name_valid(name: str) -> bool:
    """Check if name is a valid slug."""
    return name is not None and name == get_slug(name, lowercase=False)


def generate_default_name(title: str, version: Optional[str] = None) -> str:
    """Get dataset name."""
    max_length = 24
    # For compatibility with older versions use title as name if it is valid; otherwise, use encoded title
    if is_dataset_name_valid(title):
        return title

    slug = get_slug(title)
    name = slug[:max_length]

    if version:
        max_version_length = 10
        version_slug = get_slug(version)[:max_version_length]
        name = f"{name[:-(len(version_slug) + 1)]}_{version_slug}"

    return get_slug(name)


class Url:
    """Represents a schema URL reference."""

    def __init__(
        self,
        *,
        id: Optional[str] = None,
        url: Optional[Union[str, Dict[str, str]]] = None,
        url_str: Optional[str] = None,
        url_id: Optional[str] = None,
    ):
        self.id: str
        self.url: Union[str, Dict[str, str]]
        self.url_str: Optional[str] = url_str
        self.url_id: Optional[str] = url_id

        if not url:
            self.url = self._get_default_url()
        else:
            self.url = url
            if isinstance(self.url, dict):
                if "_id" in self.url:
                    self.url["@id"] = self.url.pop("_id")
                self.url_id = self.url["@id"]
                self.url_str = None
            elif isinstance(self.url, str):
                self.url_str = self.url
                self.url_id = None

        if not id or id.startswith("_:"):
            self.id = Url.generate_id(url_str=self.url_str, url_id=self.url_id)
        else:
            self.id = id

    def __repr__(self) -> str:
        return f"<Url {self.value}>"

    @staticmethod
    def generate_id(url_str, url_id):
        """Generate an identifier for Url."""
        url = url_str or url_id
        id = urlparse(url)._replace(scheme="").geturl().strip("/") if url else uuid4().hex
        id = quote(id, safe="/")

        return f"/urls/{id}"

    @property
    def value(self) -> str:
        """Returns the url value as string."""
        return cast(str, self.url_str or self.url_id)

    def _get_default_url(self):
        """Define default value for url field."""
        if self.url_str:
            return self.url_str
        elif self.url_id:
            return {"@id": self.url_id}
        else:
            raise NotImplementedError("Either url_id or url_str has to be set")


class DatasetTag(Persistent):
    """Represents a Tag of an instance of a dataset."""

    def __init__(
        self,
        *,
        dataset_id: Url,
        date_created: Optional[datetime] = None,
        description: Optional[str] = None,
        id: Optional[str] = None,
        name: str,
    ):
        if not id:
            id = DatasetTag.generate_id(dataset_id=dataset_id.value, name=name)

        self.dataset_id: Url = dataset_id
        self.date_created: datetime = parse_date(date_created) or local_now()
        self.description: Optional[str] = description
        self.id: str = id
        self.name: str = name

    @staticmethod
    def generate_id(dataset_id: str, name: str) -> str:
        """Define default value for id field."""
        identifier = Path(dataset_id).name
        name = quote(f"{name}@{identifier}", safe="")
        return f"/dataset-tags/{name}"


class Language(Immutable):
    """Represent a language of an object."""

    __slots__ = ("alternate_name", "name")

    def __init__(self, name: str, alternate_name: Optional[str] = None, id: Optional[str] = None):
        id = id or Language.generate_id(name)
        super().__init__(alternate_name=alternate_name, id=id, name=name)

    @staticmethod
    def generate_id(name: str) -> str:
        """Generate @id field."""
        name = quote(name, safe="")
        return f"/languages/{name}"


class ImageObject(Slots):
    """Represents a schema.org `ImageObject`."""

    __slots__ = ("content_url", "id", "position")

    id: str
    content_url: str
    position: str

    def __init__(self, *, content_url: str, id: str, position: int):
        id = get_path(id)
        super().__init__(content_url=content_url, position=position, id=id)

    @staticmethod
    def generate_id(dataset_id: str, position: int) -> str:
        """Generate @id field."""
        return f"{dataset_id}/images/{position}"

    @property
    def is_absolute(self):
        """Whether content_url is an absolute or relative url."""
        return bool(urlparse(self.content_url).netloc)


class RemoteEntity(Slots):
    """Reference to an Entity in a remote repository."""

    __slots__ = ("checksum", "id", "path", "url")

    def __init__(self, *, checksum: Optional[str], id: Optional[str] = None, path: Union[Path, str], url: str):
        super().__init__()
        self.checksum: str = checksum or NON_EXISTING_ENTITY_CHECKSUM
        self.id: str = id or RemoteEntity.generate_id(checksum=self.checksum, path=path, url=url)
        self.path: str = str(path)
        self.url: str = url

    @staticmethod
    def generate_id(checksum: str, path: Union[Path, str], url: str) -> str:
        """Generate an id."""
        parsed_url = urlparse(url)
        prefix = quote(posixpath.join(parsed_url.netloc.strip("/"), parsed_url.path.strip("/")))
        path = quote(str(path).strip("/"))
        return f"/remote-entities/{prefix}/{checksum}/{path}"

    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(other, RemoteEntity):
            return False
        return self.checksum == other.checksum and self.path == other.path and self.url == other.url

    def __hash__(self):
        return hash((self.checksum, self.path, self.url))


class DatasetFile(Slots):
    """A file in a dataset."""

    __slots__ = ("based_on", "date_added", "date_removed", "entity", "id", "is_external", "source", "linked", "size")

    @deal.ensure(lambda self, *_, result, **kwargs: self.date_removed is None or self.date_removed >= self.date_added)
    def __init__(
        self,
        *,
        entity: "Entity",
        based_on: Optional[RemoteEntity] = None,
        date_added: Optional[datetime] = None,
        date_removed: Optional[datetime] = None,
        id: Optional[str] = None,
        is_external: Optional[bool] = False,
        linked: Optional[bool] = False,
        source: Optional[Union[Path, str]] = None,
        size: Optional[int] = None,
    ):
        from renku.domain_model.entity import Entity

        assert entity is None or isinstance(entity, Entity), f"Invalid entity type: '{entity}'"

        super().__init__()

        self.based_on: Optional[RemoteEntity] = based_on
        self.date_added: datetime = fix_datetime(date_added) or local_now()
        self.date_removed: Optional[datetime] = fix_datetime(date_removed)
        self.entity: "Entity" = entity
        self.id: str = id or DatasetFile.generate_id()
        self.is_external: bool = is_external or False
        self.linked: bool = linked or False
        self.source: Optional[str] = str(source)
        self.size: Optional[int] = size

    @classmethod
    def from_path(
        cls,
        path: Union[str, Path],
        source=None,
        based_on: Optional[RemoteEntity] = None,
        checksum: Optional[str] = None,
        size: Optional[int] = None,
    ) -> "DatasetFile":
        """Return an instance from a path."""
        from renku.domain_model.entity import Entity

        # NOTE: Data is added from an external storage and isn't pulled yet
        if based_on and not (project_context.path / path).exists():
            checksum = based_on.checksum if based_on.checksum else NON_EXISTING_ENTITY_CHECKSUM
            id = Entity.generate_id(checksum=checksum, path=path)
            entity = Entity(id=id, checksum=checksum, path=path)
        else:
            entity = get_entity_from_revision(
                repository=project_context.repository, path=path, bypass_cache=True, checksum=checksum
            )

        is_external = False
        linked = is_linked_file(path=path, project_path=project_context.path)
        return cls(entity=entity, is_external=is_external, source=source, based_on=based_on, linked=linked, size=size)

    @staticmethod
    def generate_id():
        """Generate an identifier for DatasetFile.

        NOTE: ID should not rely on Entity properties because the same Entity can be added and removed multiple times.
        So, it should be marked by different DatasetFiles.
        """
        return f"/dataset-files/{uuid4().hex}"

    @classmethod
    def from_dataset_file(cls, other: "DatasetFile") -> "DatasetFile":
        """Return a copy with a different id."""
        self = other.copy()
        self.id = DatasetFile.generate_id()

        return self

    def __repr__(self) -> str:
        return f"<DatasetFile {self.entity.path}>"

    def correct_linked_attribute(self):
        """Replace ``is_external`` attribute with ``linked`` for linked dataset files."""
        if self.is_external and is_linked_file(self.entity.path, project_path=project_context.path):
            self.linked = True
            self.is_external = False
        elif not hasattr(self, "linked"):
            self.linked = False

    def copy(self) -> "DatasetFile":
        """Return a clone of this object."""
        return copy.copy(self)

    def is_equal_to(self, other: "DatasetFile"):
        """Compare content.

        NOTE: id is generated randomly and should not be included in this comparison.
        """
        return (
            self.based_on == other.based_on
            and self.date_added == other.date_added
            and self.date_removed == other.date_removed
            and self.entity == other.entity
            and self.is_external == other.is_external
            and self.linked == other.linked
            and self.source == other.source
        )

    @deal.ensure(lambda self, *_, result, **kwargs: self.date_removed >= self.date_added)
    def remove(self, date: Optional[datetime] = None):
        """Create a new instance and mark it as removed."""
        date_removed = fix_datetime(date) or local_now()
        self.date_removed = date_removed

    def is_removed(self) -> bool:
        """Return true if dataset is removed and should not be accessed."""
        return self.date_removed is not None

    def has_valid_checksum(self) -> bool:
        """Return if file has a valid checksum."""
        return (
            bool(self.entity.checksum)
            and self.entity.checksum != NON_EXISTING_ENTITY_CHECKSUM
            and (
                self.based_on is None
                or self.based_on.checksum != NON_EXISTING_ENTITY_CHECKSUM
                or bool(self.based_on.checksum)
            )
        )

    def has_valid_size(self) -> bool:
        """Return if file has a valid size."""
        return self.size is not None


class Dataset(Persistent):
    """Represent a dataset."""

    date_modified: Optional[datetime] = None  # type: ignore
    storage: Optional[str] = None
    datadir: Optional[str] = None

    @deal.ensure(
        lambda self, *_, result, **kwargs: (self.date_created is not None and self.date_published is None)
        or (self.date_created is None and self.date_published is not None)
    )
    @deal.ensure(
        lambda self, *_, result, **kwargs: (self.date_created is not None and self.date_modified >= self.date_created)
        or (self.date_published is not None and self.date_modified >= self.date_published)
    )
    @deal.ensure(
        lambda self, *_, result, **kwargs: self.date_removed is None or self.date_removed >= self.date_modified
    )
    def __init__(
        self,
        *,
        annotations: Optional[List["Annotation"]] = None,
        creators: Optional[List["Person"]] = None,
        datadir: Optional[Path] = None,
        dataset_files: Optional[List[DatasetFile]] = None,
        date_created: Optional[datetime] = None,
        date_published: Optional[datetime] = None,
        date_removed: Optional[datetime] = None,
        date_modified: Optional[datetime] = None,
        derived_from: Optional[Url] = None,
        description: Optional[str] = None,
        id: Optional[str] = None,
        identifier: Optional[str] = None,
        images: Optional[List[ImageObject]] = None,
        in_language: Optional[Language] = None,
        initial_identifier: Optional[str] = None,
        keywords: Optional[List[str]] = None,
        license: Optional[str] = None,
        name: Optional[str] = None,
        project_id: Optional[str] = None,
        same_as: Optional[Url] = None,
        storage: Optional[str] = None,
        title: Optional[str] = None,
        version: Optional[str] = None,
    ):
        if not name:
            assert title, "Either 'name' or 'title' must be set."
            name = generate_default_name(title, version)

        self._validate_name(name)
        self._validate_creator(creators)

        # if `date_published` is set, we are probably dealing with an imported dataset so `date_created` is not needed
        if date_published:
            date_created = None
        else:
            date_created = fix_datetime(date_created) or local_now()
        if initial_identifier is None:
            assert identifier is None, "Initial identifier can be None only when creating a new Dataset."
            initial_identifier = identifier = uuid4().hex

        self.identifier = identifier or uuid4().hex
        self.id = id or Dataset.generate_id(identifier=self.identifier)
        self.name: str = name

        self.creators: List["Person"] = creators or []
        # `dataset_files` includes existing files and those that have been removed in the previous version
        self.dataset_files: List[DatasetFile] = dataset_files or []
        self.date_created: Optional[datetime] = date_created
        self.date_modified: datetime = date_modified or local_now()
        self.date_published: Optional[datetime] = fix_datetime(date_published)
        self.date_removed: Optional[datetime] = fix_datetime(date_removed)
        self.derived_from: Optional[Url] = derived_from
        self.description: Optional[str] = description
        self.images: List[ImageObject] = images or []
        self.in_language: Optional[Language] = in_language
        self.initial_identifier: str = initial_identifier
        self.keywords: List[str] = keywords or []
        self.license: Optional[str] = license
        self.project_id: Optional[str] = project_id
        self.same_as: Optional[Url] = same_as
        self.storage: Optional[str] = storage
        self.title: Optional[str] = title
        self.version: Optional[str] = version
        self.annotations: List["Annotation"] = annotations or []

        if datadir:
            self.datadir: Optional[str] = str(datadir)

        self.correct_linked_files()

    def __setstate__(self, state):
        super().__setstate__(state)
        self.correct_linked_files()

    def correct_linked_files(self):
        """Fix linked dataset files."""
        for file in self.dataset_files:
            file.correct_linked_attribute()

    @staticmethod
    def generate_id(identifier: str) -> str:
        """Generate an identifier for Dataset."""
        return f"/datasets/{identifier}"

    @staticmethod
    def _validate_name(name):
        if not is_dataset_name_valid(name):
            raise errors.ParameterError(f"Invalid dataset name: '{name}'")

    @staticmethod
    def _validate_creator(creators):
        from renku.domain_model.provenance.agent import Person, SoftwareAgent

        creators = creators or []
        for creator in creators:
            if not isinstance(creator, (Person, SoftwareAgent)):
                raise ValueError(f"Invalid creator type: {creator}")

    @property
    def files(self) -> List[DatasetFile]:
        """Return list of existing files."""
        return [f for f in self.dataset_files if not f.is_removed()]

    @property
    def creators_csv(self):
        """Comma-separated list of creators associated with dataset."""
        return ", ".join(creator.name for creator in self.creators)

    @property
    def creators_full_csv(self):
        """Comma-separated list of creators with full identity."""
        return ", ".join(creator.full_identity for creator in self.creators)

    @property
    def keywords_csv(self):
        """Comma-separated list of keywords associated with dataset."""
        return ", ".join(self.keywords or [])

    def get_datadir(self) -> Path:
        """Return dataset's data directory relative to project's root."""
        if self.datadir:
            return Path(self.datadir)

        return Path(os.path.join(project_context.datadir, self.name))

    def __repr__(self) -> str:
        return f"<Dataset {self.identifier} {self.name}>"

    def is_derivation(self) -> bool:
        """Return if a dataset has correct derived_from."""
        return self.derived_from is not None and not self.same_as and self.id != self.derived_from.url_id

    def copy(self) -> "Dataset":
        """Return a clone of this dataset."""
        try:
            self.unfreeze()
            dataset = copy.copy(self)
        finally:
            self.freeze()

        dataset.annotations = [a.copy() for a in self.annotations]
        dataset.creators = self.creators.copy()
        dataset.dataset_files = [f.copy() for f in self.dataset_files]
        dataset.images = list(dataset.images or [])
        dataset.keywords = list(dataset.keywords or [])
        return dataset

    def replace_identifier(self, identifier: Optional[str] = None):
        """Replace dataset's identifier and update relevant fields.

        NOTE: Call this only for newly-created/-imported datasets that don't have a mutability chain because it sets
        `initial_identifier`.
        """
        assert self.derived_from is None, (
            f"Replacing identifier of dataset '{self.name}:{self.identifier}' "
            f"that is derived from {self.derived_from.url_id}"
        )

        self._assign_new_identifier(identifier)
        # NOTE: Do not unset `same_as` because it can be set for imported datasets

    @deal.ensure(
        lambda self, *_, result, **kwargs: (self.date_created is not None and self.date_published is None)
        or (self.date_created is None and self.date_published is not None)
    )
    @deal.ensure(
        lambda self, *_, result, **kwargs: (self.date_created is not None and self.date_modified >= self.date_created)
        or (self.date_published is not None and self.date_modified >= self.date_published)
    )
    @deal.ensure(lambda self, *_, result, **kwargs: self.derived_from is not None)
    def derive_from(
        self,
        dataset: "Dataset",
        creator: Optional["Person"],
        identifier: Optional[str] = None,
        date_created: Optional[datetime] = None,
    ):
        """Make `self` a derivative of `dataset` and update related fields."""
        assert dataset is not None, "Cannot derive from None"
        assert self is not dataset, f"Cannot derive from the same dataset '{self.name}:{self.identifier}'"
        assert not identifier or self.id != identifier, f"Cannot derive from the same id '{self.name}:{identifier}'"

        self._assign_new_identifier(identifier)
        # NOTE: Setting `initial_identifier` is required for migration of broken projects
        self.initial_identifier = dataset.initial_identifier
        self.derived_from = Url(url_id=dataset.id)
        self.same_as = None
        self.date_created = date_created or dataset.date_created
        self.date_modified = local_now()
        self.date_published = dataset.date_published

        if creator and hasattr(creator, "email") and not any(c for c in self.creators if c.email == creator.email):
            self.creators.append(creator)

    def _assign_new_identifier(self, identifier: Optional[str]):
        identifier = identifier or uuid4().hex
        self.initial_identifier = identifier
        self.identifier = identifier
        self.id = Dataset.generate_id(identifier)
        # NOTE: We also need to re-assign the _p_oid since identifier has changed
        self.reassign_oid()

    @deal.ensure(lambda self, *_, result, **kwargs: self.date_removed >= self.date_modified)
    def remove(self, date: Optional[datetime] = None):
        """Mark the dataset as removed."""
        self.date_removed = fix_datetime(date) or local_now()

    def is_removed(self) -> bool:
        """Return true if dataset is removed."""
        return self.date_removed is not None

    def find_file(self, path: Union[Path, str]) -> Optional[DatasetFile]:
        """Find a file in the dataset using its relative path."""
        path = str(path)
        for file in self.dataset_files:
            if str(file.entity.path) == path and not file.is_removed():
                return file

        return None

    def update_files_from(self, current_dataset: "Dataset", date: Optional[datetime] = None):
        """Check `current_files` to reuse existing entries and mark removed files."""
        new_files: Dict[str, DatasetFile] = {f.entity.path: f for f in self.files}
        current_files: Dict[str, DatasetFile] = {f.entity.path: f for f in current_dataset.files}

        files = []

        for path, file in new_files.items():
            # Use existing entries from `current_files` to avoid creating new ids
            current_file = current_files.pop(path, None)
            if current_file and file.is_equal_to(current_file):
                files.append(current_file)
            else:
                files.append(file)

        # NOTE: Whatever remains in `current_files` are removed in the newer version
        for removed_file in current_files.values():
            removed_file = DatasetFile.from_dataset_file(removed_file)
            removed_file.remove(date)
            files.append(removed_file)

        self.dataset_files = files

    def update_metadata_from(self, other: "Dataset", exclude=None):
        """Update metadata from another dataset."""
        updatable_fields = [
            "creators",
            "date_created",
            "date_published",
            "derived_from",
            "description",
            "images",
            "in_language",
            "keywords",
            "license",
            "same_as",
            "title",
            "version",
        ]
        for name in updatable_fields:
            value = getattr(other, name)
            if exclude and name in exclude:
                continue
            setattr(self, name, value)

        if self.date_published is not None:
            self.date_created = None

    def update_metadata(self, **kwargs):
        """Updates metadata."""
        editable_attributes = ["creators", "description", "keywords", "title"]
        for name, value in kwargs.items():
            if name not in editable_attributes:
                raise errors.ParameterError(f"Cannot edit field: '{name}'")
            if value is not NO_VALUE and value != getattr(self, name):
                setattr(self, name, value)

    def unlink_file(self, path, missing_ok=False) -> Optional[DatasetFile]:
        """Mark a file as removed using its relative path."""
        assert not self.immutable, f"Dataset is immutable {self}"

        file = self.find_file(path)

        if not file:
            if not missing_ok:
                raise errors.InvalidFileOperation(f"File cannot be found: {path}")
            return None

        file.remove()

        return file

    def is_within_datadir(self, path: Union[Path, str]) -> bool:
        """Return True if a given path is inside dataset's data directory."""
        datadir = get_absolute_path(self.get_datadir())
        absolute_path = get_absolute_path(path)
        return os.path.commonpath([absolute_path, datadir]) == datadir

    def add_or_update_files(self, files: Union[DatasetFile, List[DatasetFile]]):
        """Add new files or update existing files."""
        assert not self.immutable, f"Dataset is immutable {self}"

        if isinstance(files, DatasetFile):
            files = [files]

        new_files = []

        for file in cast(List[DatasetFile], files):
            existing_file = self.find_file(file.entity.path)
            if not existing_file:
                new_files.append(file)
            elif file.entity.checksum != existing_file.entity.checksum or file.date_added != existing_file.date_added:
                self.dataset_files.remove(existing_file)
                new_files.append(file)

        if not new_files:
            return

        self.dataset_files += new_files
        self._p_changed = True

    def clear_files(self):
        """Remove all files."""
        self.dataset_files = []


class DatasetCreatorsJson(marshmallow.Schema):
    """Schema for the dataset creators."""

    name = marshmallow.fields.String()
    email = marshmallow.fields.String()
    affiliation = marshmallow.fields.String()


class AnnotationJson(marshmallow.Schema):
    """Schema for Annotations."""

    source = marshmallow.fields.String()
    body = marshmallow.fields.Dict()


class DatasetDetailsJson(marshmallow.Schema):
    """Serialize a dataset to a response object."""

    name = marshmallow.fields.String(required=True)
    version = marshmallow.fields.String(allow_none=True)
    created_at = marshmallow.fields.String(allow_none=True, attribute="date_created")

    title = marshmallow.fields.String()
    creators = marshmallow.fields.List(marshmallow.fields.Nested(DatasetCreatorsJson))
    description = marshmallow.fields.String()
    keywords = marshmallow.fields.List(marshmallow.fields.String())
    identifier = marshmallow.fields.String()
    storage = marshmallow.fields.String()

    annotations = marshmallow.fields.List(marshmallow.fields.Nested(AnnotationJson))

    data_directory = marshmallow.fields.Method("get_datadir")

    @staticmethod
    def get_datadir(obj):
        """Get data directory."""
        if isinstance(obj, dict):
            return str(obj.get("datadir_path", obj.get("datadir", "")))
        if hasattr(obj, "datadir_path"):
            return obj.datadir_path

        return str(obj.get_datadir())


class DatasetFileDetailsJson(marshmallow.Schema):
    """Serialize dataset files to a response object."""

    path = marshmallow.fields.String()
    created = marshmallow.fields.DateTime()
    added = marshmallow.fields.DateTime()

    size = marshmallow.fields.String()
    is_lfs = marshmallow.fields.Boolean()

    dataset_id = marshmallow.fields.String()
    dataset_name = marshmallow.fields.String()

    creators = marshmallow.fields.List(marshmallow.fields.Nested(DatasetCreatorsJson))


class ImageObjectJson(marshmallow.Schema):
    """ImageObject json schema."""

    content_url = marshmallow.fields.String()
    position = marshmallow.fields.Integer()


class ImageObjectRequestJson(marshmallow.Schema):
    """ImageObject json schema."""

    file_id = marshmallow.fields.String()
    content_url = marshmallow.fields.String()
    position = marshmallow.fields.Integer()
    mirror_locally = marshmallow.fields.Bool(dump_default=False)


def get_file_path_in_dataset(dataset: Dataset, dataset_file: DatasetFile) -> Path:
    """Return path of a file relative to dataset's data dir."""
    try:
        return (project_context.path / dataset_file.entity.path).relative_to(
            project_context.path / dataset.get_datadir()
        )
    except ValueError:  # NOTE: File is not in the dataset's data dir
        return Path(dataset_file.entity.path)



#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service cache views for v1."""
from dataclasses import asdict

from flask import request

from renku.core.errors import AuthenticationError, ProjectNotFound
from renku.ui.service.controllers.cache_migrate_project import MigrateProjectCtrl
from renku.ui.service.controllers.cache_migrations_check import MigrationsCheckCtrl
from renku.ui.service.gateways.gitlab_api_provider import GitlabAPIProvider
from renku.ui.service.serializers.v1.cache import ProjectMigrateResponseRPC_1_0, ProjectMigrationCheckResponseRPC_1_5
from renku.ui.service.views import result_response
from renku.ui.service.views.api_versions import V1_0, V1_1, V1_2, V1_3, V1_4, V1_5
from renku.ui.service.views.decorators import accepts_json, optional_identity, requires_cache, requires_identity
from renku.ui.service.views.error_handlers import (
    handle_common_except,
    handle_migration_read_errors,
    handle_migration_write_errors,
)


@handle_common_except
@handle_migration_write_errors
@accepts_json
@requires_cache
@requires_identity
def migrate_project_view_1_0(user_data, cache):
    """
    Migrate a project.

    ---
    post:
      description: Migrate a project.
      requestBody:
        content:
          application/json:
            schema: ProjectMigrateRequest
      responses:
        200:
          description: Migration status.
          content:
            application/json:
              schema: ProjectMigrateResponseRPC_1_0
      tags:
        - cache
    """
    ctrl = MigrateProjectCtrl(cache, user_data, dict(request.json))  # type: ignore
    ctrl.RESPONSE_SERIALIZER = ProjectMigrateResponseRPC_1_0()  # type: ignore
    return ctrl.to_response()


@handle_common_except
@handle_migration_read_errors
@requires_cache
@optional_identity
def migration_check_project_view_1_5(user_data, cache):
    """
    Retrieve migration information for a project.

    ---
    get:
      description: Retrieve migration information for a project.
      parameters:
        - in: query
          schema: ProjectMigrationCheckRequest
      responses:
        200:
          description: Information about required migrations for the project.
          content:
            application/json:
              schema: ProjectMigrationCheckResponseRPC_1_5
      tags:
        - cache
    """

    from flask import jsonify

    from renku.ui.service.serializers.rpc import JsonRPCResponse
    from renku.ui.service.views.error_handlers import pretty_print_error

    ctrl = MigrationsCheckCtrl(cache, user_data, dict(request.args), GitlabAPIProvider())

    if "project_id" in ctrl.context:  # type: ignore
        result = ctrl.execute_op()
    else:
        # NOTE: use quick flow but fallback to regular flow in case of unexpected exceptions
        try:
            result = ctrl._fast_op_without_cache()
        except (AuthenticationError, ProjectNotFound):
            raise
        except BaseException:
            result = ctrl.execute_op()

    if isinstance(result.core_compatibility_status, Exception):
        return jsonify(JsonRPCResponse().dump({"error": pretty_print_error(result.core_compatibility_status)}))

    if isinstance(result.template_status, Exception):
        return jsonify(JsonRPCResponse().dump({"error": pretty_print_error(result.template_status)}))

    if isinstance(result.dockerfile_renku_status, Exception):
        return jsonify(JsonRPCResponse().dump({"error": pretty_print_error(result.dockerfile_renku_status)}))

    return result_response(ProjectMigrationCheckResponseRPC_1_5(), asdict(result))


def add_v1_specific_endpoints(cache_blueprint):
    """Add v1 only endpoints to blueprint."""
    cache_blueprint.route("/cache.migrate", methods=["POST"], provide_automatic_options=False, versions=[V1_0])(
        migrate_project_view_1_0
    )
    cache_blueprint.route(
        "/cache.migrations_check",
        methods=["GET"],
        provide_automatic_options=False,
        versions=[V1_0, V1_1, V1_2, V1_3, V1_4, V1_5],
    )(migration_check_project_view_1_5)
    return cache_blueprint



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helpers functions for metadata management/parsing."""

import fnmatch
import os
import re
import tempfile
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Tuple, Union

from packaging.version import Version

from renku.core import errors
from renku.core.config import get_value, set_value
from renku.core.constant import POINTERS, RENKU_HOME, RENKU_PROTECTED_PATHS, RENKU_TMP
from renku.core.migration.utils import OLD_METADATA_PATH
from renku.core.util import communication
from renku.core.util.os import is_subpath

if TYPE_CHECKING:
    from renku.core.dataset.providers.api import ProviderCredentials
    from renku.domain_model.provenance.agent import Person


def construct_creators(creators: List[Union[dict, str]], ignore_email=False) -> Tuple[List["Person"], List[str]]:
    """Parse input and return a list of Person."""
    creators = creators or []

    people = []
    no_email_warnings = []
    for creator in creators:
        person, no_email_warning = construct_creator(creator, ignore_email=ignore_email)

        if person:
            people.append(person)

        if no_email_warning:
            no_email_warnings.append(no_email_warning)

    return people, no_email_warnings


def construct_creator(creator: Union[dict, str], ignore_email) -> Tuple[Optional["Person"], Optional[str]]:
    """Parse input and return an instance of Person."""
    from renku.domain_model.provenance.agent import Person

    if not creator:
        return None, None

    if isinstance(creator, str):
        person = Person.from_string(creator)
    elif isinstance(creator, dict):
        person = Person.from_dict(creator)
    else:
        raise errors.ParameterError("Invalid creator type")

    message = 'A valid format is "Name <email> [affiliation]"'

    if not person.name:  # pragma: no cover
        raise errors.ParameterError(f'Name is invalid: "{creator}".\n{message}')

    if not person.email:
        if not ignore_email:  # pragma: no cover
            raise errors.ParameterError(f'Email is invalid: "{creator}".\n{message}')
        else:
            no_email_warning = str(creator)
    else:
        no_email_warning = None

    return person, no_email_warning


def is_external_file(path: Union[Path, str], project_path: Path):
    """Checks if a path is an external file."""
    path = project_path / path
    if not path.is_symlink() or not is_subpath(path=path, base=project_path):
        return False

    pointer = os.readlink(path)
    return str(os.path.join(RENKU_HOME, POINTERS)) in pointer


def is_linked_file(path: Union[Path, str], project_path: Path) -> bool:
    """Return True if a dataset file is a linked file."""
    path = project_path / path
    if not path.is_symlink() or not is_subpath(path=path.resolve(), base=project_path):
        return False

    pointer = os.readlink(path)
    return os.path.join(RENKU_HOME, POINTERS) in pointer


def read_renku_version_from_dockerfile(path: Optional[Union[Path, str]] = None) -> Optional[Version]:
    """Read RENKU_VERSION from the content of path if a valid version is available."""
    from renku.domain_model.project_context import project_context

    path = Path(path) if path else project_context.dockerfile_path
    if not path.exists():
        return None

    m = re.search(r"^\s*ARG RENKU_VERSION=(\d+\.\d+\.\d+\S*)$", path.read_text(), flags=re.MULTILINE)
    if not m:
        return None

    try:
        return Version(m.group(1))
    except ValueError:
        return None


def replace_renku_version_in_dockerfile(dockerfile_content: str, version: str) -> str:
    """Replace Renku version in the Dockerfile."""
    return re.sub(
        r"^\s*ARG RENKU_VERSION=(\d+\.\d+\.\d+\S*)$",
        f"ARG RENKU_VERSION={version}",
        dockerfile_content,
        flags=re.MULTILINE,
    )


def make_project_temp_dir(project_path: Path) -> Path:
    """Create a temporary directory inside project's temp path."""
    base = project_path / RENKU_HOME / RENKU_TMP
    base.mkdir(parents=True, exist_ok=True)

    return Path(tempfile.mkdtemp(dir=base))


def store_credentials(section: str, key: str, value: str) -> None:
    """Write provider's credentials."""
    section, key = _get_section_and_key(section=section, key=key)
    set_value(section=section, key=key, value=value, global_only=True)


def read_credentials(section: str, key: str) -> Optional[str]:
    """Read provider's credentials."""
    section, key = _get_section_and_key(section=section, key=key)
    return get_value(section=section, key=key)


def get_canonical_key(key: str) -> str:
    """Make a consistent configuration key."""
    return key.replace(" ", "-").lower()


def _get_section_and_key(section: str, key: str) -> Tuple[str, str]:
    section = "http" if section.lower() == "renku" else section

    return section, get_canonical_key(key)


def prompt_for_credentials(provider_credentials: "ProviderCredentials") -> None:
    """Prompt for provider credentials if needed and update and store them."""
    if not provider_credentials:
        return

    provider_credentials.read()

    prompt_to_store = False
    for key in provider_credentials.get_credentials_names_with_no_value():
        prompt_to_store = True
        value = communication.prompt(f"Enter a value for '{key}'", type=str, default="")

        provider_credentials[key] = value

    if prompt_to_store:
        if communication.confirm("Store credentials?", default=True):
            provider_credentials.store()


def is_protected_path(path: Path) -> bool:
    """Checks if a path is a protected path."""
    from renku.domain_model.project_context import project_context

    try:
        path_in_repo = str(path.relative_to(project_context.path))
    except ValueError:
        return False

    for protected_path in RENKU_PROTECTED_PATHS:
        if fnmatch.fnmatch(path_in_repo, protected_path):
            return True

    return False


def is_renku_project() -> bool:
    """Check if repository is a renku project."""
    from renku.domain_model.project_context import project_context

    try:
        return project_context.project is not None
    except ValueError:  # NOTE: Error in loading due to an older schema
        return project_context.metadata_path.joinpath(OLD_METADATA_PATH).exists()



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""CompositePlan view model."""

import json
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Dict, List, NamedTuple, Optional, Union

from renku.command.view_model.agent import PersonViewModel
from renku.core.errors import ParameterNotFoundError
from renku.domain_model.workflow.composite_plan import CompositePlan
from renku.domain_model.workflow.parameter import (
    CommandInput,
    CommandOutput,
    CommandParameter,
    CommandParameterBase,
    ParameterLink,
    ParameterMapping,
)
from renku.domain_model.workflow.plan import AbstractPlan

if TYPE_CHECKING:
    from renku.command.view_model.plan import PlanViewModel


class ParameterReference(NamedTuple):
    """Reference to a workflow parameter."""

    id: str
    plan_id: str
    type: str
    name: str


parameter_type_mapping: Dict[type, str] = {
    CommandInput: "Input",
    CommandOutput: "Output",
    CommandParameter: "Parameter",
    ParameterMapping: "Mapping",
}


def _parameter_id_to_plan_id(parameter: CommandParameterBase, parent_plan: AbstractPlan, latest: bool = False):
    """Extract plan id from a parameter id."""
    from renku.core.workflow.plan import get_latest_plan

    plan_path = parent_plan.get_parameter_path(parameter)

    if plan_path is None or len(plan_path) < 1:
        raise ParameterNotFoundError(parameter.name, parent_plan.name)

    containing_plan = plan_path[-1]

    if latest:
        containing_plan = get_latest_plan(containing_plan)
    return containing_plan.id


def _parameter_to_type_string(parameter: CommandParameterBase):
    """Get a type string for a parameter."""
    return parameter_type_mapping[type(parameter)]


class ParameterMappingViewModel:
    """View model for ``ParameterMapping``."""

    def __init__(
        self,
        name: str,
        default_value: str,
        maps_to: List[str],
        targets: List[ParameterReference],
        description: Optional[str] = None,
        plan_id: Optional[str] = None,
    ):
        self.name = name
        self.default_value = default_value
        self.maps_to = maps_to
        self.targets = targets
        self.description = description
        self.plan_id = plan_id
        self.type = "Mapping"

    @classmethod
    def from_mapping(cls, mapping: ParameterMapping, plan: AbstractPlan, latest: bool = False):
        """Create view model from ``ParameterMapping``.

        Args:
            mapping(ParameterMapping): Mapping to create view model from.
            plan(AbstractPlan): Parent plan.
            latest(bool): Whether to get latest plan data.

        Returns:
            View model of mapping.
        """
        targets = [
            ParameterReference(
                t.id, _parameter_id_to_plan_id(t, plan, latest=latest), _parameter_to_type_string(t), t.name
            )
            for t in mapping.mapped_parameters
        ]
        return cls(
            name=mapping.name,
            default_value=str(mapping.default_value),
            maps_to=[t.name for t in mapping.mapped_parameters],
            targets=targets,
            description=mapping.description,
            plan_id=plan.id,
        )


class ParameterLinkViewModel:
    """View model for ``ParameterLink``."""

    def __init__(
        self,
        id: str,
        plan_id: str,
        source: str,
        sinks: List[str],
        source_entry: Optional[ParameterReference] = None,
        sink_entries: Optional[List[ParameterReference]] = None,
    ):
        self.id = id
        self.plan_id = plan_id
        self.source = source
        self.sinks = sinks
        self.source_entry = source_entry
        self.sink_entries = sink_entries
        self.type = "Link"

    @classmethod
    def from_link(cls, link: ParameterLink, plan: AbstractPlan, latest: bool = False):
        """Create view model from ``ParameterLink``.

        Args:
            link(ParameterLink): Link to get view model from.
            plan(AbstractPlan): Parent plan.
            latest(bool): Whether to get latest plan data.

        Returns:
            View model for link.
        """
        source_path = plan.get_parameter_path(link.source)
        source_path.append(link.source)
        source_path = ".".join(p.name for p in source_path[1:])

        source_entry = ParameterReference(
            link.source.id,
            _parameter_id_to_plan_id(link.source, plan, latest=latest),
            _parameter_to_type_string(link.source),
            link.source.name,
        )

        sinks = []
        sink_entries = []

        for sink in link.sinks:
            sink_path = plan.get_parameter_path(sink)
            sink_path.append(sink)
            sink_path = ".".join(p.name for p in sink_path[1:])
            sinks.append(sink_path)
            sink_entries.append(
                ParameterReference(
                    sink.id,
                    _parameter_id_to_plan_id(sink, plan, latest=latest),
                    _parameter_to_type_string(sink),
                    sink.name,
                )
            )
        return cls(
            id=link.id,
            plan_id=plan.id,
            source=source_path,
            sinks=sinks,
            source_entry=source_entry,
            sink_entries=sink_entries,
        )


class StepViewModel:
    """View model for ``Plan`` inside ``CompositePlan``."""

    def __init__(self, id: str, name: str):
        self.id = id
        self.name = name


class CompositePlanViewModel:
    """A view model for a ``CompositePlan``."""

    def __init__(
        self,
        id: str,
        name: str,
        created: datetime,
        mappings: List[ParameterMappingViewModel],
        links: List[ParameterLinkViewModel],
        steps: List[Union["CompositePlanViewModel", "PlanViewModel"]],
        keywords: List[str],
        description: Optional[str] = None,
        creators: Optional[List[PersonViewModel]] = None,
        annotations: Optional[str] = None,
        touches_existing_files: Optional[bool] = None,
        latest: Optional[str] = None,
        duration: Optional[timedelta] = None,
    ):
        self.id = id
        self.name = name
        self.description = description
        self.created = created
        self.mappings = mappings
        self.links = links
        self.steps = steps
        self.creators = creators
        self.keywords = keywords
        self.annotations = annotations
        self.touches_existing_files = touches_existing_files
        self.latest = latest
        self.type = "CompositePlan"
        self.full_command = ""

        if duration is not None:
            self.duration = duration.seconds

    @classmethod
    def from_composite_plan(cls, plan: CompositePlan, latest: bool = False):
        """Create view model from ``Plan``.

        Args:
            plan(CompositePlan): Composite Plan to get view model from.
            latest(bool): Whether to get latest plan data.

        Returns:
            View model of composite Plan.
        """
        from renku.command.view_model.plan import plan_view

        return cls(
            id=plan.id,
            name=plan.name,
            description=plan.description,
            created=plan.date_created,
            mappings=[ParameterMappingViewModel.from_mapping(mapping, plan, latest) for mapping in plan.mappings],
            links=[ParameterLinkViewModel.from_link(link, plan, latest) for link in plan.links],
            steps=[plan_view(p) for p in getattr(plan, "newest_plans", plan.plans)],
            creators=[PersonViewModel.from_person(p) for p in plan.creators] if plan.creators else None,
            keywords=plan.keywords,
            annotations=json.dumps([{"id": a.id, "body": a.body, "source": a.source} for a in plan.annotations])
            if plan.annotations
            else None,
            latest=getattr(plan, "latest", None),
            touches_existing_files=getattr(plan, "touches_existing_files", False),
            duration=getattr(plan, "duration", None),
        )



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common functionality for data providers."""

import re
from pathlib import Path
from typing import List

from renku.core import errors
from renku.core.dataset.providers.api import StorageProviderInterface
from renku.core.dataset.providers.models import DatasetAddAction, DatasetAddMetadata
from renku.domain_model.dataset import RemoteEntity
from renku.domain_model.project_context import project_context


def get_metadata(
    provider: StorageProviderInterface, uri: str, destination: Path, dataset_add_action: DatasetAddAction
) -> List["DatasetAddMetadata"]:
    """Add files from a URI to a dataset."""
    if re.search(r"[*?]", uri):
        raise errors.ParameterError("Wildcards like '*' or '?' are not supported for cloud storage URIs.")

    storage = provider.get_storage()

    destination_path_in_repo = Path(destination).relative_to(project_context.repository.path)
    hashes = storage.get_hashes(uri=uri)
    if dataset_add_action == DatasetAddAction.NONE:
        dataset_add_action = DatasetAddAction.REMOTE_STORAGE
    return [
        DatasetAddMetadata(
            entity_path=destination_path_in_repo / hash.path,
            url=hash.uri,
            action=dataset_add_action,
            based_on=RemoteEntity(checksum=hash.hash if hash.hash else "", url=hash.uri, path=hash.path),
            source=Path(hash.uri),
            destination=destination / hash.path,
            provider=provider,
            size=hash.size,
        )
        for hash in hashes
    ]



#
# Copyright 2017-2023 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku generic database gateway implementation."""

from pathlib import Path
from typing import Generator, List, Union

import BTrees
from persistent import Persistent
from persistent.list import PersistentList
from zc.relation.catalog import Catalog
from zc.relation.queryfactory import TransposingTransitive
from zope.interface import Attribute, Interface, implementer

from renku.core.interface.database_gateway import IDatabaseGateway
from renku.domain_model.dataset import Dataset
from renku.domain_model.project_context import project_context
from renku.domain_model.provenance.activity import Activity, ActivityCollection
from renku.domain_model.workflow.plan import AbstractPlan
from renku.infrastructure.database import RenkuOOBTree


class IActivityDownstreamRelation(Interface):
    """Interface for activity downstream relation."""

    downstream = Attribute("the downstream activities")
    upstream = Attribute("the upstream activities")


@implementer(IActivityDownstreamRelation)
class ActivityDownstreamRelation:
    """Implementation of Downstream interface."""

    def __init__(self, downstream, upstream):
        self.downstream = downstream
        self.upstream = upstream

        self.id = f"{upstream.id}:{downstream.id}"

    def __repr__(self):
        return f"<ActivityDownstreamRelation {self.id} at 0x{id(self):0x}>"


def dump_activity(activity: Activity, catalog, cache) -> str:
    """Get storage token for an activity."""
    return activity.id


def load_activity(token: str, catalog, cache) -> Activity:
    """Load activity from storage token."""
    database = project_context.database
    return database["activities"].get(token)


def dump_downstream_relations(relation: ActivityDownstreamRelation, catalog, cache):
    """Dump relation entry to database."""
    btree = project_context.database["_downstream_relations"]

    btree[relation.id] = relation

    return relation.id


def load_downstream_relations(token, catalog, cache):
    """Load relation entry from database."""
    btree = project_context.database["_downstream_relations"]

    return btree[token]


def initialize_database(database):
    """Initialize an empty database with all required metadata."""
    # NOTE: A list of existing and removed activities
    database.add_index(name="activities", object_type=Activity, attribute="id")
    database.add_root_object(name="activities-by-usage", obj=RenkuOOBTree())
    database.add_root_object(name="activities-by-generation", obj=RenkuOOBTree())

    database.add_index(name="activity-collections", object_type=ActivityCollection, attribute="id")

    database.add_root_object(name="_downstream_relations", obj=RenkuOOBTree())

    activity_catalog = Catalog(dump_downstream_relations, load_downstream_relations, btree=BTrees.family32.OO)
    activity_catalog.addValueIndex(
        IActivityDownstreamRelation["downstream"],  # type: ignore[misc]
        dump_activity,
        load_activity,
        btree=BTrees.family32.OO,
    )
    activity_catalog.addValueIndex(
        IActivityDownstreamRelation["upstream"],  # type: ignore[misc]
        dump_activity,
        load_activity,
        btree=BTrees.family32.OO,
    )
    # NOTE: Transitive query factory is needed for transitive (follow more than 1 edge) queries
    downstream_transitive_factory = TransposingTransitive("downstream", "upstream")
    activity_catalog.addDefaultQueryFactory(downstream_transitive_factory)

    database.add_root_object(name="activity-catalog", obj=activity_catalog)

    database.add_index(name="plans", object_type=AbstractPlan, attribute="id")
    database.add_index(name="plans-by-name", object_type=AbstractPlan, attribute="name")

    database.add_index(name="datasets", object_type=Dataset, attribute="name")
    database.add_index(name="datasets-provenance-tails", object_type=Dataset, attribute="id")
    database.add_index(name="datasets-tags", object_type=PersistentList)


class DatabaseGateway(IDatabaseGateway):
    """Gateway for base database operations."""

    def initialize(self) -> None:
        """Initialize the database."""
        database = project_context.database

        database.clear()
        initialize_database(database)
        database.commit()

    def commit(self) -> None:
        """Commit changes to database."""
        database = project_context.database

        database.commit()

    def get_modified_objects_from_revision(self, revision_or_range: str) -> Generator[Persistent, None, None]:
        """Get all database objects modified in a revision."""
        repository = project_context.repository

        if ".." in revision_or_range:
            commits: Union[Generator, List] = repository.iterate_commits(revision=revision_or_range)
        else:
            commits = [repository.get_commit(revision_or_range)]

        for commit in commits:
            for file in commit.get_changes(f"{project_context.database_path}/**"):
                if file.deleted:
                    continue

                oid = Path(file.a_path).name

                yield project_context.database.get(oid)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Azure dataset provider."""

import urllib
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Tuple, cast

from renku.core import errors
from renku.core.dataset.providers.api import (
    AddProviderInterface,
    ProviderApi,
    ProviderCredentials,
    ProviderPriority,
    StorageProviderInterface,
)
from renku.core.dataset.providers.common import get_metadata
from renku.core.dataset.providers.models import DatasetAddAction
from renku.core.interface.storage import IStorage
from renku.core.util.metadata import get_canonical_key, prompt_for_credentials
from renku.core.util.urls import get_scheme
from renku.domain_model.project_context import project_context
from renku.infrastructure.storage.factory import StorageFactory

if TYPE_CHECKING:
    from renku.core.dataset.providers.models import DatasetAddMetadata
    from renku.domain_model.dataset import Dataset


class AzureProvider(ProviderApi, StorageProviderInterface, AddProviderInterface):
    """Azure provider."""

    priority = ProviderPriority.HIGHEST
    name = "Azure"
    is_remote = True

    def __init__(self, uri: str):
        super().__init__(uri=uri)

        account, endpoint, container, _ = parse_azure_uri(uri=self.uri)

        self._account: str = account
        self._endpoint: str = endpoint
        self._container = container

    @staticmethod
    def supports(uri: str) -> bool:
        """Whether or not this provider supports a given URI."""
        return get_scheme(uri) == "azure"

    @staticmethod
    def supports_storage(uri: str) -> bool:
        """Whether or not this provider supports a given URI storage."""
        return AzureProvider.supports(uri=uri)

    def get_metadata(
        self, uri: str, destination: Path, dataset_add_action: DatasetAddAction = DatasetAddAction.NONE, **kwargs
    ) -> List["DatasetAddMetadata"]:
        """Get metadata of files that will be added to a dataset."""
        return get_metadata(provider=self, uri=uri, destination=destination, dataset_add_action=dataset_add_action)

    def convert_to_storage_uri(self, uri: str) -> str:
        """Convert backend-specific URI to a URI that is usable by the IStorage implementation."""
        _, _, container, path = parse_azure_uri(uri=uri)
        return f"azure://{container}/{path}"

    def get_credentials(self) -> "AzureCredentials":
        """Return an instance of provider's credential class."""
        return AzureCredentials(provider=self)

    def get_storage(self, credentials: Optional["ProviderCredentials"] = None) -> "IStorage":
        """Return the storage manager for the provider."""
        azure_configuration = {
            "type": "azureblob",
        }

        if not credentials:
            credentials = self.get_credentials()
            prompt_for_credentials(credentials)

        return StorageFactory.get_storage(
            storage_scheme="azure",
            provider=self,
            credentials=credentials,
            configuration=azure_configuration,
        )

    @property
    def account(self) -> str:
        """Azure account name."""
        return self._account

    @property
    def endpoint(self) -> str:
        """Return Azure container endpoint."""
        return self._endpoint

    @property
    def container(self) -> str:
        """Return Azure container name."""
        return self._container

    def on_create(self, dataset: "Dataset") -> None:
        """Hook to perform provider-specific actions on a newly-created dataset."""
        credentials = self.get_credentials()
        prompt_for_credentials(credentials)
        storage = self.get_storage(credentials=credentials)

        # NOTE: The underlying rclone tool cannot tell if a directory within an Azure container exists or not
        if not storage.exists(self.uri):
            raise errors.ParameterError(f"Azure container '{self.container}' doesn't exists.")

        project_context.repository.add_ignored_pattern(pattern=str(dataset.get_datadir()))


class AzureCredentials(ProviderCredentials):
    """Azure-specific credentials."""

    def __init__(self, provider: AzureProvider):
        super().__init__(provider=provider)

        # NOTE: Set account name so that users don't need to re-enter it
        self.data[get_canonical_key("Account")] = self.provider.account

    @staticmethod
    def get_credentials_names() -> Tuple[str, ...]:
        """Return a tuple of the required credentials for a provider."""
        return "Account", "Key"

    @property
    def provider(self) -> AzureProvider:
        """Return the associated provider instance."""
        return cast(AzureProvider, self._provider)

    def get_credentials_section_name(self) -> str:
        """Get section name for storing credentials.

        NOTE: This methods should be overridden by subclasses to allow multiple credentials per providers if needed.
        """
        return f"{self.provider.account}.{self.provider.endpoint}"


def parse_azure_uri(uri: str) -> Tuple[str, str, str, str]:
    """Extract account, endpoint, container, and path within the container from a given URI.

    NOTE: We support azure://<account-name>.<endpoint>/<container-name>/<path> or
    azure://<account-name>/<container-name>/<path>.
    """
    parsed_uri = urllib.parse.urlparse(uri)

    if parsed_uri.scheme.lower() != "azure":
        raise errors.ParameterError(
            f"Invalid scheme in Azure URI: {uri}.\n"
            "Valid format is 'azure://<account-name>/<container-name>/<path>' or "
            "'azure://<account-name>.<endpoint>/<container-name>/<path>'",
            show_prefix=False,
        )

    account, _, endpoint = parsed_uri.netloc.partition(".")
    path = parsed_uri.path.strip("/")
    container, _, path = path.partition("/")

    if not account or not container:
        raise errors.ParameterError(
            f"Missing account and/or container name in Azure URI: {uri}.\n"
            "Valid format is 'azure://<account-name>/<container-name>/<path>' or "
            "'azure://<account-name>.<endpoint>/<container-name>/<path>'",
            show_prefix=False,
        )

    endpoint = endpoint.lower() or "blob.core.windows.net"

    return account, endpoint, container, path.strip("/")



#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service view decorators."""
from functools import wraps

from flask import request
from marshmallow import ValidationError

from renku.ui.service.cache import cache
from renku.ui.service.errors import ProgramContentTypeError, UserAnonymousError
from renku.ui.service.serializers.headers import OptionalIdentityHeaders, RequiredIdentityHeaders
from renku.ui.service.views.error_handlers import handle_redis_except


def requires_identity(f):
    """Wrapper which indicates that route requires user identification."""

    @wraps(f)
    def decorated_function(*args, **kws):
        """Represents decorated function."""
        try:
            user_identity = RequiredIdentityHeaders().load(request.headers)
        except (ValidationError, KeyError) as e:
            raise UserAnonymousError(e)

        return f(user_identity, *args, **kws)

    return decorated_function


def optional_identity(f):
    """Wrapper which indicates partial dependency on user identification."""

    @wraps(f)
    def decorated_function(*args, **kws):
        """Represents decorated function."""
        user_identity = OptionalIdentityHeaders().load(request.headers)
        return f(user_identity, *args, **kws)

    return decorated_function


@handle_redis_except
def requires_cache(f):
    """Wrapper which injects cache object into view."""

    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        return f(cache, *args, **kwargs)

    return decorated_function


def accepts_json(f):
    """Wrapper which ensures only JSON payload can be in request."""

    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        content_type = "application/json"
        wrong_type = False
        if "Content-Type" not in request.headers:
            wrong_type = True
        else:
            header_check = request.headers["Content-Type"] == content_type
            if not request.is_json or not header_check:
                wrong_type = True

        if wrong_type:
            raise ProgramContentTypeError(request.headers.get("Content-Type"), content_type)

        return f(*args, **kwargs)

    return decorated_function



#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service datasets edit controller."""
from typing import Dict, List, Union, cast

from renku.command.dataset import edit_dataset_command
from renku.core.dataset.request_model import ImageRequestModel
from renku.core.util.metadata import construct_creators
from renku.domain_model.constant import NO_VALUE, NoValueType
from renku.domain_model.provenance.agent import Person
from renku.ui.service.cache.models.job import Job
from renku.ui.service.config import CACHE_UPLOADS_PATH, MESSAGE_PREFIX
from renku.ui.service.controllers.api.abstract import ServiceCtrl
from renku.ui.service.controllers.api.mixins import RenkuOpSyncMixin
from renku.ui.service.controllers.utils.datasets import set_url_for_uploaded_images
from renku.ui.service.serializers.datasets import DatasetEditRequest, DatasetEditResponseRPC
from renku.ui.service.views import result_response


class DatasetsEditCtrl(ServiceCtrl, RenkuOpSyncMixin):
    """Controller for datasets edit endpoint."""

    REQUEST_SERIALIZER = DatasetEditRequest()
    RESPONSE_SERIALIZER = DatasetEditResponseRPC()

    def __init__(self, cache, user_data, request_data, migrate_project=False):
        """Construct a datasets edit list controller."""
        self.ctx = cast(Dict, DatasetsEditCtrl.REQUEST_SERIALIZER.load(request_data))
        self.ctx["commit_message"] = f"{MESSAGE_PREFIX} dataset edit {self.ctx['name']}"

        super().__init__(cache, user_data, request_data, migrate_project=migrate_project)

    @property
    def context(self):
        """Controller operation context."""
        return self.ctx

    def renku_op(self):
        """Renku operation for the controller."""
        warnings: List[Union[str, Dict]] = []

        if "images" in self.ctx:
            images = self.ctx.get("images") or []

            user_cache_dir = CACHE_UPLOADS_PATH / self.user.user_id

            set_url_for_uploaded_images(images=images, cache=self.cache, user=self.user)

            images = [
                ImageRequestModel(
                    content_url=img["content_url"],
                    position=img["position"],
                    mirror_locally=img.get("mirror_locally", False),
                    safe_image_paths=[user_cache_dir],
                )
                for img in images
            ]
        else:
            images = NO_VALUE

        creators: Union[NoValueType, List[Person]]
        if "creators" in self.ctx:
            creators, warnings = construct_creators(self.ctx.get("creators"))  # type: ignore
        else:
            creators = NO_VALUE

        if "title" in self.ctx:
            title = self.ctx.get("title")
        else:
            title = NO_VALUE

        if "description" in self.ctx:
            description = self.ctx.get("description")
        else:
            description = NO_VALUE

        if "keywords" in self.ctx:
            keywords = self.ctx.get("keywords") or []
        else:
            keywords = NO_VALUE

        if "custom_metadata" in self.ctx:
            custom_metadata = self.ctx.get("custom_metadata")
        else:
            custom_metadata = NO_VALUE

        if "custom_metadata_source" in self.ctx:
            custom_metadata_source = self.ctx.get("custom_metadata")
        else:
            custom_metadata_source = NO_VALUE

        if custom_metadata_source is NO_VALUE and custom_metadata is not NO_VALUE:
            custom_metadata_source = "renku"

        result = (
            edit_dataset_command()
            .with_commit_message(self.ctx["commit_message"])
            .build()
            .execute(
                self.ctx["name"],
                title=title,
                description=description,
                creators=creators,
                keywords=keywords,
                images=images,
                custom_metadata=custom_metadata,
                custom_metadata_source=custom_metadata_source,
            )
        )

        edited = result.output
        return edited, warnings

    def to_response(self):
        """Execute controller flow and serialize to service response."""
        op_result, remote_branch = self.execute_and_sync()

        if isinstance(op_result, Job):
            return result_response(DatasetsEditCtrl.JOB_RESPONSE_SERIALIZER, op_result)

        edited, warnings = op_result

        if "creators" in edited:
            edited["creators"] = self.ctx.get("creators")

        response = {
            "edited": edited,
            "warnings": warnings,
            "remote_branch": remote_branch,
        }

        return result_response(DatasetsEditCtrl.RESPONSE_SERIALIZER, response)
